# vim: ts=4:sts=4:sw=4
#
# @date 2019-10-21
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""This module contains methods for parsing and transforming time-related
attributes."""

import numpy
import numpy.typing
import typing
import logging
import datetime
import dateutil.relativedelta
from idf_converter.lib.types import InputOptions, CoverageInfo, Coverage
from idf_converter.lib.types import Duration, Granule

logger = logging.getLogger(__name__)
AnyDatetime = typing.Union[datetime.datetime, numpy.typing.NDArray,
                           typing.List[datetime.datetime]]


class MissingTimeCoverageRelativeBounds(ValueError):
    """Error raised when either the "time_coverage_relative_start" or the
    "time_coverage_relative_end" input options have not been specified whereas
    they are required to build sub-granule in case the input file contains
    several times."""
    pass


class InvalidDatetimeFormat(ValueError):
    """Error raised when a datetime passed as a string cannot be parsed due to
    its format."""
    def __init__(self, value: str) -> None:
        """"""
        self.invalid_string = value


class WrongTimeOffsetType(TypeError):
    """Error raised when a relative time coverage bound is neither a string nor
    a float."""
    def __init__(self, value: typing.Any) -> None:
        self.value = value


def parse_datetime(value: str, fmt: typing.Optional[str] = None
                   ) -> datetime.datetime:
    """"""
    if fmt is not None:
        try:
            dt_value = datetime.datetime.strptime(value, fmt)
            return dt_value
        except ValueError:
            raise InvalidDatetimeFormat(value)

    # Fix quirks with number of hours
    if '59:59:59' in value:
        value = value.replace('59:59:59', '23:59:59')
    if '595959' in value:
        value = value.replace('595959', '235959')

    # Fix datetimes provided with millisecond precision
    if '.' == value[-5]:
        value = value.replace('Z', '000Z')
    elif '.' == value[-4]:
        value = f'{value}000'

    # Fix UTC datetimes that include spaces before the "Z" timezone
    while value.endswith(' Z'):
        value = value.replace(' Z', 'Z')

    time_fmts = ('%Y-%m-%dT%H:%M:%S.%fZ',
                 '%Y-%m-%dT%H:%M:%SZ',
                 '%Y-%m-%dT%H:%M:%S',
                 '%Y%m%dT%H%M%S.%fZ',
                 '%Y%m%dT%H%M%SZ',
                 '%Y%m%dT%H%M%S',
                 '%Y%m%d %H%M%S.%fZ',
                 '%Y%m%d %H%M%SZ',
                 '%Y%m%d %H%M%S',
                 '%Y-%m-%d %H:%M:%S.%fZ',
                 '%Y-%m-%d %H:%M:%SZ',
                 '%Y-%m-%d %H:%M:%S',
                 '%Y%m%d%H%M%S.%fZ',
                 '%Y%m%d%H%M%SZ',
                 '%Y%m%d%H%M%S',
                 '%Y-%m-%d',
                 '%Y%m%d',
                 '%Y-%j',
                 )

    for dt_fmt in time_fmts:
        try:
            dt_value = datetime.datetime.strptime(value, dt_fmt)
            break
        except ValueError:
            pass  # silenced
    else:
        raise InvalidDatetimeFormat(value)
    return dt_value


def get_duration_from_iso8601(iso8601_str: str) -> typing.Optional[Duration]:
    """"""
    if not iso8601_str.startswith('P'):
        return None

    time_index = iso8601_str.find('T')
    year_index = iso8601_str.find('Y')
    month_index = iso8601_str[:time_index].find('M')
    day_index = iso8601_str.find('D')
    hour_index = iso8601_str.find('H', time_index)
    minute_index = iso8601_str.find('M', time_index)
    second_index = iso8601_str.find('S', time_index)

    latest_index = 0 + 1

    years = 0
    if -1 < year_index:
        years = int(iso8601_str[latest_index:year_index])
        latest_index = year_index + 1

    months = 0
    if -1 < month_index:
        months = int(iso8601_str[latest_index:month_index])
        latest_index = year_index + 1

    days = 0
    if -1 < day_index:
        days = int(iso8601_str[latest_index:day_index])
        latest_index = year_index + 1

    latest_index = time_index + 1

    hours = 0
    if -1 < hour_index:
        hours = int(iso8601_str[latest_index:hour_index])
        latest_index = year_index + 1

    minutes = 0
    if -1 < minute_index:
        minutes = int(iso8601_str[latest_index:minute_index])
        latest_index = year_index + 1

    seconds = 0
    if -1 < second_index:
        seconds = int(iso8601_str[latest_index:second_index])

    result = dateutil.relativedelta.relativedelta(years=years, months=months,
                                                  days=days, hours=hours,
                                                  minutes=minutes,
                                                  seconds=seconds)
    return result


def _get_time_coverage_info(input_opts: InputOptions,
                            granule: Granule) -> CoverageInfo:
    """"""
    start = granule.meta.get('time_coverage_start', None)
    end = granule.meta.get('time_coverage_end', None)

    # Relative offset for time coverage lower bound
    relative_start = input_opts.get('time_coverage_relative_start', None)
    if relative_start is not None:
        if isinstance(relative_start, str):
            if 'month_beginning' == relative_start.lower():
                relative_start = 'month_beginning'
            elif 'month_end' == relative_start.lower():
                relative_start = 'month_end'
            else:
                relative_start = float(relative_start)
        else:
            relative_start = None

    # Relative offset for time coverage upper bound
    relative_end = input_opts.get('time_coverage_relative_end', None)
    if relative_end is not None:
        if isinstance(relative_end, str):
            if 'month_beginning' == relative_end.lower():
                relative_end = 'month_beginning'
            elif 'month_end' == relative_end.lower():
                relative_end = 'month_end'
            else:
                relative_end = float(relative_end)
        else:
            relative_end = None

    return (start, end, relative_start, relative_end)


def get_time_coverages(input_opts: InputOptions,
                       granule: Granule,
                       dtimes: numpy.typing.NDArray,
                       ) -> typing.List[Coverage]:
    """"""

    coverage_info = _get_time_coverage_info(input_opts, granule)
    start, end, relative_start, relative_end = coverage_info
    has_offsets = (None not in (relative_start, relative_end))

    if 1 >= len(dtimes):
        if (start is not None) and (end is not None) and not has_offsets:
            # Boundaries are already defined and relative bounds (higher
            # priority)  have not been specified so there is nothing else to do
            return [(start, end), ]

    # From this point on, relative bounds are mandatory
    if has_offsets is False:
        logger.error('time_coverage_relative_start and '
                     'time_coverage_relative_end must be provided in '
                     'order to extract granules at several times from '
                     'a single input file, or if ')
        raise MissingTimeCoverageRelativeBounds()

    bounds: typing.List[Coverage] = []
    for i in range(0, dtimes.size):
        # Might happen with forecast data that allocate enough entries for
        # their final form but or not complete yet. For example a forecast
        # at hourly time resolution, distributed as one file per day: the
        # size of the time dimension can be 24 even if all time slots are
        # not filled at first because the model will fill them eventually,
        # in future runs.
        if (dtimes[i] is None) or (dtimes[i] is numpy.ma.masked):
            logger.warning(f'{i+1}th time is invalid and has been skipped')
            bounds.append((None, None))
            continue

        if 'month_beginning' == relative_start:
            start_dt = datetime.datetime(dtimes[i].year, dtimes[i].month, 1)
        elif 'month_end' == relative_start:
            one_month = dateutil.relativedelta.relativedelta(months=1)
            one_microsecond = datetime.timedelta(microseconds=1)
            _start_dt = datetime.datetime(dtimes[i].year, dtimes[i].month, 1)
            start_dt = _start_dt + one_month - one_microsecond
        else:
            time_offset = None
            if isinstance(relative_start, str):
                time_offset = float(relative_start)
            elif isinstance(relative_start, float):
                time_offset = relative_start
            else:
                raise WrongTimeOffsetType(relative_start)
            start_dt = dtimes[i] + datetime.timedelta(seconds=time_offset)

        if 'month_beginning' == relative_end:
            stop_dt = datetime.datetime(dtimes[i].year, dtimes[i].month, 1)
        elif 'month_end' == relative_end:
            one_month = dateutil.relativedelta.relativedelta(months=1)
            one_microsecond = datetime.timedelta(microseconds=1)
            _stop_dt = datetime.datetime(dtimes[i].year, dtimes[i].month, 1)
            stop_dt = _stop_dt + one_month - one_microsecond
        else:
            time_offset = None
            if isinstance(relative_end, str):
                time_offset = float(relative_end)
            elif isinstance(relative_end, float):
                time_offset = relative_end
            else:
                raise WrongTimeOffsetType(relative_end)
            stop_dt = dtimes[i] + datetime.timedelta(seconds=time_offset)

        bounds.append((start_dt, stop_dt,))
    return bounds


def as_datetimes_array(dtimes: AnyDatetime
                       ) -> numpy.typing.NDArray:
    """"""
    dt_cast = '_to_real_datetime'
    if hasattr(dtimes, dt_cast):
        # conversion from cftime + scalar -> 1-item array
        return numpy.array([getattr(dtimes, dt_cast)()])
    elif isinstance(dtimes, datetime.datetime):
        # scalar -> 1-item array
        return numpy.array([dtimes])

    return numpy.array([getattr(_, dt_cast)()
                        if ((_ is not None) and hasattr(_, dt_cast))
                        else _
                        for _ in dtimes])



# vim: ts=4:sts=4:sw=4
#
# @date 2019-10-01
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import os
import numpy
import numpy.typing
import typing
import netCDF4
import logging
import idf_converter.lib
import idf_converter.lib.nir_bc
from idf_converter.lib.types import InputOptions, OutputOptions, ReaderResult
from idf_converter.lib.types import Granule

logger = logging.getLogger(__name__)

MasksDict = typing.Dict[str, numpy.typing.NDArray]
SpecAngleMaskParams = typing.Tuple[typing.Optional[numpy.typing.NDArray],
                                   typing.List[str]]
DATA_MODEL = 'SWATH'

# S-2 : 665, 560, 490
# red : 650 nm
# green : 510 nm
# blue : 475 nm
# https://sentinel.esa.int/web/sentinel/user-guides/sentinel-3-olci/overview/heritage  # noqa
# https://sentinel.esa.int/web/sentinel/technical-guides/sentinel-3-olci/level-1/fr-or-rr-toa-radiances  # noqa
# Oa1   400 	15
# Oa2 	412.5 	10
# Oa3 	442.5 	10
# Oa4 	490 	10   BLUE
# Oa5 	510 	10
# Oa6 	560 	10   GREEN
# Oa7 	620 	10
# Oa8 	665 	10   RED
# Oa9 	673.75 	7.5
# Oa10 	681.25 	7.5
# Oa11 	708.75 	10
# Oa12 	753.75 	7.5
# Oa13 	761.25 	2.5
# Oa14 	764.375 3.75
# Oa15 	767.5 	2.5
# Oa16 	778.75 	15
# Oa17 	865 	20 SUN GLITTER
# Oa18 	885 	10
# Oa19 	900 	10
# Oa20 	940 	20
# Oa21 	1 020 	40

# TSM: Total Suspended Matter concentration
# CDM: Coloured Dissolved Matter
# The diffuse attenuation coefficient at 490 nm (Kd490) indicates the turbidity
# of the water column - how visible light in the blue to green region of the
# spectrum penetrates within the water column. The value of Kd490 represents
# the rate which light at 490 nm is attenuated with depth.


class InputPathMissing(Exception):
    """Error raised when the "path" input option has not been specified by the
    user."""
    pass


class GeoCoordinatesNotFound(IOError):
    """Error raised when the path for the geographical coordinates file does
    not exist."""
    def __init__(self, path: str) -> None:
        """"""
        self.path = path


class TimeCoordinatesNotFound(IOError):
    """Error raised when the path for the temporal coordinates file does not
    exist."""
    def __init__(self, path: str) -> None:
        """"""
        self.path = path


class QualityFlagsNotFound(IOError):
    """Error raised when the path for quality flags file does not exist."""
    def __init__(self, path: str) -> None:
        """"""
        self.path = path


class ReflectanceFileNotFound(IOError):
    """Error raised when the path for the file which contains reflectance
    values does not exist."""
    def __init__(self, path: str) -> None:
        """"""
        self.path = path


class DataFileNotFound(IOError):
    """Error raised when the path for the file which contains either the
    CHL_NN, CHL_OC4ME, TSM_NN or KD490_M07 values does not exist."""
    def __init__(self, path: str) -> None:
        """"""
        self.path = path


class TieGeometriesNotFound(IOError):
    """Error raised when the path for the file for tie coordinates does not
    exist."""
    def __init__(self, path: str) -> None:
        """"""
        self.path = path


class TieGeometriesDimensionMismatch(ValueError):
    """Error raised when the alongtrack dimension has different sizes in the
    tie_geometries.nc and geo_coordinates.nc files."""
    def __init__(self, tie_size: int, geo_size: int) -> None:
        """"""
        self.tie_size = tie_size
        self.geo_size = geo_size


class InvalidMinSpecularAngleValue(ValueError):
    """Error raised when the value specified by the user for the specular angle
    threshold cannot be parsed as a float number."""
    def __init__(self, value: str) -> None:
        """"""
        self.value = value


class InvalidMaskLowSpecularAngleOption(ValueError):
    """Error raised when the mask_low_specular_angle option passed by the user
    contains more than 1 colon, which is not the expected format."""
    def __init__(self, value: str) -> None:
        """"""
        self.value = value


def help() -> typing.Tuple[str, str]:
    """Describe options supported by this reader.

    Returns
    -------
    str
        Description of the supported options
    """
    inp = ('    path\tPath of the input file',)
    out = ('',)
    return ('\n'.join(inp), '\t'.join(out))


def retrieve_mask_bit(quality_path: str
                      ) -> typing.Dict[str, int]:
    handler = netCDF4.Dataset(quality_path, 'r')
    bits = handler.variables['WQSF'].flag_masks
    meaning_str = handler.variables['WQSF'].flag_meanings
    handler.close()
    meanings = meaning_str.split(' ')
    mask_bits = {}
    for iflag, ibit in zip(meanings, bits):
        mask_bits[iflag] = int(numpy.log(ibit)/numpy.log(2))
    return mask_bits


def get_contrast_common_mask(quality_path: str,
                             channels: typing.List[str],
                             mask_bits: typing.Dict[str, int],
                             ) -> numpy.typing.NDArray:
    """"""
    # We keep inland waters, whitecaps
    # Allow Hisolzen flag otherwise it removes high latitude data
    off_flags = numpy.uint32(0)
    off_flags = off_flags + numpy.uint32(1 << mask_bits['INVALID'])
    off_flags = off_flags + numpy.uint32(1 << mask_bits['LAND'])
    off_flags = off_flags + numpy.uint32(1 << mask_bits['CLOUD'])
    off_flags = off_flags + numpy.uint32(1 << mask_bits['SNOW_ICE'])
    off_flags = off_flags + numpy.uint32(1 << mask_bits['COSMETIC'])
    off_flags = off_flags + numpy.uint32(1 << mask_bits['SUSPECT'])
    off_flags = off_flags + numpy.uint32(1 << mask_bits['SATURATED'])
    off_flags = off_flags + numpy.uint32(1 << mask_bits['ADJAC'])
    off_flags = off_flags + numpy.uint32(1 << mask_bits['WV_FAIL'])
    off_flags = off_flags + numpy.uint32(1 << mask_bits['PAR_FAIL'])
    off_flags = off_flags + numpy.uint32(1 << mask_bits['AC_FAIL'])
    off_flags = off_flags + numpy.uint32(1 << mask_bits['KDM_FAIL'])
    off_flags = off_flags + numpy.uint32(1 << mask_bits['CLOUD_MARGIN'])
    off_flags = off_flags + numpy.uint32(1 << mask_bits['CLOUD_AMBIGUOUS'])

    # Read quality flags
    quality_handler = netCDF4.Dataset(quality_path, 'r')
    _quality_flags = idf_converter.lib.extract_variable_values(quality_handler,
                                                               'WQSF')
    quality_flags = numpy.ma.array(_quality_flags)
    quality_handler.close()

    contrast_mask = numpy.zeros(quality_flags.shape, dtype='bool')
    contrast_mask = (contrast_mask |
                     (numpy.bitwise_and(quality_flags, off_flags) > 0))

    return contrast_mask


def get_contrast_bands_mask(quality_path: str,
                            targets: typing.List[str],
                            mask_bits: typing.Dict[str, int]) -> MasksDict:
    """"""
    # Read quality flags
    quality_handler = netCDF4.Dataset(quality_path, 'r')
    _quality_flags = idf_converter.lib.extract_variable_values(quality_handler,
                                                               'WQSF')
    quality_flags = numpy.ma.array(_quality_flags)
    quality_handler.close()

    masks = {}
    for var_id in targets:
        contrast_mask = numpy.zeros(quality_flags.shape, dtype='bool')

        if 'CHL_NN' == var_id:
            off_flags = numpy.uint32(0)
            off_flags = off_flags + numpy.uint32(1 << mask_bits['OCNN_FAIL'])
            contrast_mask = (contrast_mask |
                             (numpy.bitwise_and(quality_flags, off_flags) > 0))
        elif 'CHL_OC4ME' == var_id:
            off_flags = numpy.uint32(0)
            off_flags = off_flags + numpy.uint32(1 << mask_bits['OC4ME_FAIL'])
            contrast_mask = (contrast_mask |
                             (numpy.bitwise_and(quality_flags, off_flags) > 0))

        masks[var_id] = contrast_mask
    return masks


def get_dynamic_bands_mask(granule: Granule,
                           targets: typing.List[str]) -> MasksDict:
    """"""
    masks = {}
    for var_id in targets:
        var_data = granule.vars[var_id]['array']

        dyn_mask = numpy.ma.getmaskarray(var_data)
        if var_id.startswith('Oa'):
            # Mask null and negative values: they are inferior or equal to
            # atmospheric correction and should probably have been flagged as
            # clouds.
            # They should therefore not be used to computed contrast for ocean
            dyn_mask = (dyn_mask | (var_data <= 0))
        masks[var_id] = dyn_mask
    return masks


def compute_specular_angle_mask(input_opts: InputOptions,
                                granule: Granule,
                                input_dir: str,
                                ) -> SpecAngleMaskParams:
    """"""
    # Default parameters
    min_angle = 9.0
    targets = ['CHL_NN', 'CHL_OC4ME', 'KD490', 'KD490_M07']

    # Check if mask has been requested
    _mask_low_spec_angle = input_opts.get('mask_low_specular_angle', 'false')

    opts_count = _mask_low_spec_angle.count(':')
    if 0 >= opts_count:
        # if mask_neede is true, default parameters will be used
        mask_needed = _mask_low_spec_angle.lower() in ('yes', 'true', '1')
    elif 1 == opts_count:
        mask_needed = True

        _min_angle, _targets = _mask_low_spec_angle.split(':')

        try:
            min_angle = float(_min_angle.strip())
        except ValueError:
            raise InvalidMinSpecularAngleValue(_min_angle)

        targets = [_.strip() for _ in _targets.split(',')
                   if 0 < len(_.strip())]
    else:
        raise InvalidMaskLowSpecularAngleOption(_mask_low_spec_angle)

    if mask_needed is False:
        return (None, [])

    # Compute specular angle
    data_angle_path = os.path.join(input_dir, 'tie_geometries.nc')
    if not os.path.exists(data_angle_path):
        raise TieGeometriesNotFound(data_angle_path)

    data_angle = netCDF4.Dataset(data_angle_path, 'r')
    tie_sza = data_angle.variables['SZA'][:]
    tie_saa = data_angle.variables['SAA'][:]
    tie_oza = data_angle.variables['OZA'][:]
    tie_oaa = data_angle.variables['OAA'][:]
    tie_ac_factor = data_angle.ac_subsampling_factor
    data_angle.close()

    nrow = granule.dims['row']
    ncell = granule.dims['cell']

    if tie_sza.shape[0] != nrow:
        logger.error('Wrong along track dimension in tie_geometries.nc')
        raise TieGeometriesDimensionMismatch(tie_sza.shape[0], nrow)

    tie_sza = tie_sza[:, :].astype('float32')
    tie_saa = tie_saa[:, :].astype('float32')
    tie_oza = tie_oza[:, :].astype('float32')
    tie_oaa = tie_oaa[:, :].astype('float32')
    tie_zx, tie_zy = idf_converter.lib.nir_bc.compute_tilt(tie_sza, tie_saa,
                                                           tie_oza, tie_oaa)
    tie_z = numpy.sqrt(tie_zx ** 2. + tie_zy ** 2.)

    ac_index = numpy.arange(0, ncell, dtype='float32')
    z = idf_converter.lib.nir_bc.interp_tie_ac([tie_z, ],
                                               tie_ac_factor, ac_index)[0]
    z_angle = numpy.rad2deg(numpy.arctan(z))

    # Mask data when specular angle is too low
    low_specular_angle_mask = (z_angle < min_angle)

    return (low_specular_angle_mask, targets)


def read_data(input_opts: InputOptions,
              output_opts: OutputOptions
              ) -> typing.Iterator[ReaderResult]:
    """Read input file, extract data and metadata, store them in a Granule
    object then prepare formatting instructions to finalize the conversion to
    IDF format.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule, list)
        A tuple which contains four elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the extracted information
          has been stored

        - a :obj:list of :obj:dict describing the formatting operations that
          the converter must perform before serializing the result in IDF
          format
    """
    variables = {'Oa04_reflectance': {'name': 'Oa04',
                                      'units': '',
                                      'valid_min': -0.02,
                                      'valid_max': 0.02,
                                      'options': {}},
                 'Oa06_reflectance': {'name': 'Oa06',
                                      'units': '',
                                      'valid_min': -0.02,
                                      'valid_max': 0.02,
                                      'options': {}},
                 'Oa09_reflectance': {'name': 'Oa09',
                                      'units': '',
                                      'valid_min': -0.02,
                                      'valid_max': 0.02,
                                      'options': {}},
                 'Oa17_reflectance': {'name': 'Oa17',
                                      'units': '',
                                      'valid_min': -0.02,
                                      'valid_max': 0.02,
                                      'options': {}},
                 'CHL_NN': {'name': 'CHL_NN',
                            'valid_min': -2.100,
                            'valid_max': 1.350,
                            'options': {}},
                 'CHL_OC4ME': {'name': 'CHL_OC4ME',
                               'valid_min': -1.700,
                               'valid_max': 1.550,
                               'options': {}},
                 'TSM_NN': {'name': 'TSM_NN',
                            'valid_min': -1.500,
                            'valid_max': 1.000,
                            'options': {}},
                 'KD490_M07': {'name': 'KD490',
                               'valid_min': -1.500,
                               'valid_max': 0.500,
                               'options': {}}}

    channels = list(variables.keys())
    visible_vars = ('Oa04_reflectance', 'Oa06_reflectance', 'Oa09_reflectance')
    infrared_vars = ('Oa17_reflectance',)
    reflectance_vars = list(visible_vars) + list(infrared_vars)
    non_reflectance_vars = ('CHL_NN', 'CHL_OC4ME', 'TSM_NN', 'KD490_M07')

    idf_version = output_opts.get('idf_version', '1.0')
    granule = idf_converter.lib.create_granule(idf_version, DATA_MODEL)
    granule.vars = variables

    _input_dir = input_opts.get('path', None)
    if _input_dir is None:
        raise InputPathMissing()
    input_dir = os.path.normpath(_input_dir)

    # Check existance of required files
    geo_path = os.path.join(input_dir, 'geo_coordinates.nc')
    if not os.path.exists(geo_path):
        raise GeoCoordinatesNotFound(geo_path)

    time_path = os.path.join(input_dir, 'time_coordinates.nc')
    if not os.path.exists(time_path):
        raise TimeCoordinatesNotFound(time_path)

    quality_path = os.path.join(input_dir, 'wqsf.nc')
    if not os.path.exists(quality_path):
        raise QualityFlagsNotFound(quality_path)

    # Read variables
    for var_id in reflectance_vars:
        file_name = f'{var_id}.nc'
        file_path = os.path.join(input_dir, file_name)
        if not os.path.exists(file_path):
            raise ReflectanceFileNotFound(file_path)
        f_handler = netCDF4.Dataset(file_path, 'r')
        idf_converter.lib.extract_variable_attributes(f_handler, var_id,
                                                      granule)
        band = idf_converter.lib.extract_variable_values(f_handler, var_id)
        granule.vars[var_id]['array'] = numpy.ma.array(band)
        granule.vars[var_id]['datatype'] = band.dtype
        f_handler.close()

    for var_id in ('CHL_NN', 'CHL_OC4ME', 'TSM_NN'):
        file_name = f'{var_id.lower()}.nc'
        file_path = os.path.join(input_dir, file_name)
        if not os.path.exists(file_path):
            raise DataFileNotFound(file_path)
        f_handler = netCDF4.Dataset(file_path, 'r')
        idf_converter.lib.extract_variable_attributes(f_handler, var_id,
                                                      granule)
        band = idf_converter.lib.extract_variable_values(f_handler, var_id)
        granule.vars[var_id]['array'] = numpy.ma.array(band)
        granule.vars[var_id]['datatype'] = band.dtype
        f_handler.close()

    var_id = 'KD490_M07'
    file_name = 'trsp.nc'
    file_path = os.path.join(input_dir, file_name)
    if not os.path.exists(file_path):
        raise DataFileNotFound(file_path)
    f_handler = netCDF4.Dataset(file_path, 'r')
    idf_converter.lib.extract_variable_attributes(f_handler, var_id, granule)
    band = idf_converter.lib.extract_variable_values(f_handler, var_id)
    granule.vars[var_id]['array'] = numpy.ma.array(band)
    granule.vars[var_id]['datatype'] = band.dtype
    f_handler.close()

    # Extract geo coordinates information
    geo_handler = netCDF4.Dataset(geo_path, 'r')
    idf_converter.lib.extract_global_attributes(geo_handler, input_opts,
                                                granule)
    if 'title' in granule.meta.keys():
        granule.meta['title'] = granule.meta['title'].rsplit(',', 1)[0]
    nrow = geo_handler.dimensions['rows'].size
    ncell = geo_handler.dimensions['columns'].size
    lon = idf_converter.lib.extract_variable_values(geo_handler, 'longitude')
    lon = numpy.ma.array(lon)
    lat = idf_converter.lib.extract_variable_values(geo_handler, 'latitude')
    lat = numpy.ma.array(lat)
    geo_handler.close()

    granule.dims['row'] = nrow
    granule.dims['cell'] = ncell

    # Handle longitude continuity
    dlon = lon[1:, :] - lon[:-1, :]
    if 180.0 <= numpy.max(numpy.abs(dlon)):
        lon[numpy.where(lon < 0.0)] = lon[numpy.where(lon < 0.0)] + 360.0

    # Extract time coordinates information
    time_handler = netCDF4.Dataset(time_path, 'r')
    _time = idf_converter.lib.extract_variable_values(time_handler,
                                                      'time_stamp')
    start_timestamp = _time[0]
    end_timestamp = _time[-1]
    timestamp_units = time_handler.variables['time_stamp'].units
    time_handler.close()

    granule.vars['lat'] = {'array': lat,
                           'units': 'degrees_north',
                           'datatype': lat.dtype,
                           'options': {}}
    granule.vars['lon'] = {'array': lon,
                           'units': 'degrees_east',
                           'datatype': lon.dtype,
                           'options': {}}

    # Format time information
    start_dt = netCDF4.num2date(start_timestamp, timestamp_units)
    stop_dt = netCDF4.num2date(end_timestamp, timestamp_units)

    granule_name = os.path.basename(input_dir)

    granule.meta['idf_granule_id'] = granule_name
    granule.meta['idf_subsampling_factor'] = 0
    granule.meta['idf_spatial_resolution'] = 300
    granule.meta['idf_spatial_resolution_units'] = 'm'
    granule.meta['time_coverage_start'] = start_dt
    granule.meta['time_coverage_end'] = stop_dt
    granule.meta['institution'] = 'ESA'
    granule.meta['platform'] = 'Sentinel-3'
    granule.meta['platform_type'] = 'leo satellite'
    granule.meta['sensor'] = 'OLCI'
    granule.meta['sensor_type'] = 'medium-resolution imaging spectrometer'

    # Add transforms
    transforms = []

    # Mask pixels at low specular angle if requested by user
    spec_angle_mask_params = compute_specular_angle_mask(input_opts, granule,
                                                         input_dir)
    spec_angle_mask, _spec_angle_mask_targets = spec_angle_mask_params
    spec_angle_mask_targets = [_ for _ in _spec_angle_mask_targets
                               if _ in granule.vars.keys()]
    if (spec_angle_mask is not None) and (0 < len(spec_angle_mask_targets)):
        transforms.append(('static_common_mask',
                           {'targets': spec_angle_mask_targets,
                            'mask': spec_angle_mask}))

    # Compute extrema depending on valid data proportion for reflectance bands
    # Use default values if there is not enough valid data
    # Use percentiles otherwise
    # Bands other than reflectance ones keep the default min/max
    mask_bits = retrieve_mask_bit(quality_path)
    ignored_common_mask = get_contrast_common_mask(quality_path, channels,
                                                   mask_bits)
    transforms.append(('contrast_from_pct',
                       {'targets': reflectance_vars,
                        'valid_threshold': 0.001,  # 0.1%
                        'common_mask': ignored_common_mask,
                        'bands_mask': None,
                        'dynamic_bands_mask': get_dynamic_bands_mask,
                        'min_percentile': 1.0,
                        'max_percentile': 90.0}))
    # TODO: add clipping for NIR
    # _min = numpy.clip(_min, 1.5, 10.0)
    # _max = numpy.clip(_max, 30.0, 60.0)

    # APPLY MEDIAN FILTER
    # only for non-reflectance bands
    kernels = {'CHL_NN': [3, 3],
               'CHL_OC4ME': [3, 3],
               'TSM_NN': [3, 3],
               'KD490_M07': [3, 3]}
    transforms.append(('median_filter',
                       {'targets': non_reflectance_vars,
                        'kernels': kernels}))

    # For non-reflectance bands, apply common mask
    transforms.append(('static_common_mask',
                       {'targets': non_reflectance_vars,
                        'mask': ignored_common_mask}))

    # For non-reflectance bands, also apply band-specific masks (Chl-a bands)
    ignored_bands_mask = get_contrast_bands_mask(quality_path, channels,
                                                 mask_bits)
    transforms.append(('static_bands_mask',
                       {'targets': non_reflectance_vars,
                        'masks': ignored_bands_mask}))

    output_opts['__export'] = channels

    yield input_opts, output_opts, granule, transforms



# vim: ts=4:sts=4:sw=4
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

import typing

__description__ = ('The idf_converter package provides tools to convert Earth '
                   'Observation data into IDF files.')
__version__ = '0.1'
__author__ = 'Lucile Gaultier, Gilles Guitton, Sylvain Herlédan'
__author_email__ = 'idf-support@oceandatalab.com'
__url__ = 'https://seascope.oceandatalab.com'
__keywords__: typing.List[str] = []
__classifiers__ = ['Development Status :: 5 - Production/Stable',
                   'Environment :: Console',
                   'Intended Audience :: Science/Research',
                   'License :: OSI Approved :: GNU Lesser General Public '
                   'License v3 or later (LGPLv3+)',
                   'Natural Language :: English',
                   'Programming Language :: Python :: 3',
                   'Programming Language :: Python :: 3.7',
                   'Programming Language :: Python :: Implementation :: '
                   'CPython',
                   'Topic :: Scientific/Engineering :: GIS',
                   'Typing :: Typed', ]



# vim: ts=4:sts=4:sw=4
#
# @date 2019-09-25
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""This module provides methods to manipulate the numeric scale and adapt
extrema.
"""

import numpy
import numpy.typing
import typing
import logging
from idf_converter.lib.types import InputOptions, OutputOptions, Granule
from idf_converter.lib.types import FormatterResult, NumValuesDict, Extrema

logger = logging.getLogger(__name__)
MaskGetter = typing.Callable[[Granule, typing.Iterable[str]], NumValuesDict]
ExtremaMethod = typing.Callable[[typing.Dict[str, typing.Any], float, float],
                                Extrema]


def logscale(input_opts: InputOptions,
             output_opts: OutputOptions,
             granule: Granule,
             *args: typing.Iterable[typing.Any],
             base: float,
             targets: typing.Iterable[str]) -> FormatterResult:
    """
    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format
    granule: idf_converter.lib.Granule
        Object representing the granule data and metadata
    base: float
        Base for the logarithm. Natural log will be used if the base passed as
        argument is <= 0.
    targets: tuple
        Identifers for variables that must be converted to logarithmic scale

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule)
        A tuple which contains three elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the targetted variables
          have been converted to logarithmic scale
    """
    ln_base = 1  # base e
    if 0 < base:
        ln_base = numpy.log(base)

    for var_id in targets:
        lin_values = granule.vars[var_id]['array']
        values_mask = numpy.ma.getmaskarray(lin_values)
        if values_mask.all():
            logger.warning(f'Log scale not applied to {var_id} because all '
                           'values are masked')
            continue
        nan_mask = numpy.isnan(lin_values)
        lin_values[numpy.where(nan_mask)] = 1  # will be overwritten later
        negative_mask = (lin_values <= 0)
        invalid_mask = (negative_mask | values_mask)
        log_values = numpy.log(numpy.ma.getdata(lin_values),
                               where=(~invalid_mask)) / ln_base
        log_values[numpy.where(nan_mask)] = numpy.nan

        log_vmin = numpy.log(granule.vars[var_id]['valid_min']) / ln_base
        log_vmax = numpy.log(granule.vars[var_id]['valid_max']) / ln_base

        # use log(vmin) where values were negative
        log_values[numpy.where(negative_mask)] = log_vmin

        log_values[numpy.where(values_mask)] = numpy.nan

        granule.vars[var_id]['array'] = log_values
        granule.vars[var_id]['valid_min'] = log_vmin
        granule.vars[var_id]['valid_max'] = log_vmax

        if 'units' in granule.vars[var_id].keys():
            _units = granule.vars[var_id]['units']
            if 0 < base:
                granule.vars[var_id]['units'] = f'log{base}({_units})'
            else:
                granule.vars[var_id]['units'] = f'log({_units})'

        comment = ('Logarithm applied to improve contrast and avoid '
                   'sacrificing either small or large scale structures due to '
                   'data storage resolution limitations')
        if 'comment' in granule.vars[var_id].keys():
            _comment = granule.vars[var_id]['comment']
            granule.vars[var_id]['comment'] = f'{_comment}\n{comment}'
        else:
            granule.vars[var_id]['comment'] = comment

    return (input_opts, output_opts, granule)


def stretch_extent(input_opts: InputOptions,
                   output_opts: OutputOptions,
                   granule: Granule,
                   *args: typing.Iterable[typing.Any],
                   targets: typing.Iterable[str],
                   stretch_factor: float,
                   change_vmin: bool,
                   change_vmax: bool) -> FormatterResult:
    """"""
    for var_id in targets:
        vmin = granule.vars[var_id]['valid_min']
        vmax = granule.vars[var_id]['valid_max']

        stretched_extent = stretch_factor * (vmax - vmin)
        if change_vmin == change_vmax:
            mid_value = 0.5 * (vmin + vmax)
            vmin = mid_value - 0.5 * stretched_extent
            vmax = mid_value + 0.5 * stretched_extent
        elif change_vmax:
            vmax = vmin + stretched_extent
        elif change_vmin:
            vmin = vmax - stretched_extent

        granule.vars[var_id]['valid_min'] = vmin
        granule.vars[var_id]['valid_max'] = vmax

    return input_opts, output_opts, granule


def share_extrema(input_opts: InputOptions,
                  output_opts: OutputOptions,
                  granule: Granule,
                  *args: typing.Iterable[typing.Any],
                  targets: typing.List[str]) -> FormatterResult:
    """"""
    vmins = []
    vmaxs = []
    for var_id in targets:
        vmins.append(granule.vars[var_id]['valid_min'])
        vmaxs.append(granule.vars[var_id]['valid_max'])

    shared_min: numpy.typing.ScalarType = numpy.min(vmins)
    shared_max: numpy.typing.ScalarType = numpy.max(vmaxs)
    for var_id in targets:
        granule.vars[var_id]['valid_min'] = shared_min
        granule.vars[var_id]['valid_max'] = shared_max

    return input_opts, output_opts, granule


def contrast_from_pct(input_opts: InputOptions,
                      output_opts: OutputOptions,
                      granule: Granule,
                      *args: typing.Iterable[typing.Any],
                      targets: typing.Iterable[str],
                      valid_threshold: float,
                      common_mask: typing.Optional[numpy.typing.NDArray],
                      bands_mask: typing.Optional[NumValuesDict],
                      dynamic_bands_mask: typing.Optional[MaskGetter],
                      min_percentile: float,
                      max_percentile: float) -> FormatterResult:
    """"""
    dyn_bands_mask: typing.Dict[str, numpy.typing.NDArray] = {}
    if dynamic_bands_mask is not None:
        dyn_bands_mask = dynamic_bands_mask(granule, targets)

    for var_id in targets:
        var_data = granule.vars[var_id]['array']
        invalid_mask = numpy.ma.getmaskarray(var_data)
        if common_mask is not None:
            invalid_mask = (invalid_mask | common_mask)
        if bands_mask is not None and var_id in bands_mask.keys():
            invalid_mask = (invalid_mask | bands_mask[var_id])
        if var_id in dyn_bands_mask.keys():
            invalid_mask = (invalid_mask | dyn_bands_mask[var_id])
        valid_data = numpy.ma.getdata(var_data[numpy.where(~invalid_mask)])
        valid_data_ratio = float(valid_data.size) / float(var_data.data.size)

        # If there are not enough valid data, leave the default min/max
        if valid_data_ratio <= valid_threshold:
            continue

        vmin = numpy.percentile(valid_data, min_percentile)
        vmax = numpy.percentile(valid_data, max_percentile)
        granule.vars[var_id]['valid_min'] = vmin
        granule.vars[var_id]['valid_max'] = vmax
    return input_opts, output_opts, granule


def limit_extrema(input_opts: InputOptions,
                  output_opts: OutputOptions,
                  granule: Granule,
                  *args: typing.Iterable[typing.Any],
                  targets: typing.Iterable[str],
                  min_threshold: typing.Optional[float] = None,
                  max_threshold: typing.Optional[float] = None
                  ) -> FormatterResult:
    """"""
    for var_id in targets:
        if min_threshold is not None:
            var_min = granule.vars[var_id]['valid_min']
            _min = max(var_min, min_threshold)
            granule.vars[var_id]['valid_min'] = _min
        if max_threshold is not None:
            var_max = granule.vars[var_id]['valid_max']
            _max = min(var_max, max_threshold)
            granule.vars[var_id]['valid_max'] = _max
    return input_opts, output_opts, granule


def extrema_methods(input_opts: InputOptions,
                    output_opts: OutputOptions,
                    granule: Granule,
                    *args: typing.Iterable[typing.Any],
                    targets: typing.Iterable[str],
                    methods: typing.Dict[str, ExtremaMethod],
                    min_values: typing.Dict[str, float],
                    max_values: typing.Dict[str, float]) -> FormatterResult:
    """
    Compute custom min/max for each targetted variable. The methods for
    computing the extrema  must accept an entry from the Granule.vars
    dictionary as input, as well as two float values (min and max).

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format
    granule: idf_converter.lib.Granule
        Object representing the granule data and metadata
    targets: tuple
        Identifers for variables whose extrema will be modified
    methods: dict
        Dictionary with variable identifiers as keys and methods for computing
        custom extrema as values
    min_values: dict
        Dictionary with variable identifiers as keys and floats as values, to
        be passed as argument for the min value
    max_values: dict
        Dictionary with variable identifiers as keys and floats as values, to
        be passed as argument for the max value

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule)
        A tuple which contains three elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the targetted variables
          have been masked
    """
    for var_id in targets:
        _min, _max = methods[var_id](granule.vars[var_id],
                                     min_values[var_id],
                                     max_values[var_id])
        granule.vars[var_id]['valid_min'] = _min
        granule.vars[var_id]['valid_max'] = _max

    return input_opts, output_opts, granule



# vim: ts=4:sts=4:sw=4
#
# @date 2019-09-25
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import numpy
import typing
import logging
import numpy.typing
import scipy.interpolate
from idf_converter.lib import mod360
from idf_converter.lib.types import InputOptions, OutputOptions, Granule
from idf_converter.lib.types import FormatterResult, NumValuesDict, AtmCorrData

logger = logging.getLogger(__name__)

# Scale factor used to circumvent underflow errors
UFLOW_FIX = 10**6


class IncompatibleLUTSize(ValueError):
    """Error raised when the number of entries in the LUT is not a multiple of
    (number of observation zenith angles * number of zenithal deltas)."""
    def __init__(self, lut_size: int, lut_size_factor: float) -> None:
        """"""
        self.lut_size = lut_size
        self.lut_size_factor = lut_size_factor
        _msg = ('The size of the LUT does not match the hardcoded '
                'parameters: the number of lines per sza value should '
                f'be {lut_size_factor}')
        super(IncompatibleLUTSize, self).__init__(_msg)


def interp_tie_ac(tie_values_list: typing.List[numpy.typing.NDArray],
                  tie_ac_factor: int,
                  ac_index: numpy.typing.NDArray
                  ) -> typing.List[numpy.typing.NDArray]:
    """"""
    # Linear interpolation over across-track dimension
    tie_ac_index = numpy.arange(tie_values_list[0].shape[1],
                                dtype='float32') * tie_ac_factor
    indint = numpy.clip(numpy.searchsorted(tie_ac_index, ac_index) - 1,
                        0, tie_ac_index.size - 2)
    indratio = (ac_index - tie_ac_index[indint]) / \
               (tie_ac_index[indint + 1] - tie_ac_index[indint])
    indratio = indratio[numpy.newaxis, :]
    int_values = []
    for val in tie_values_list:
        ac_diff = val[:, indint + 1] - val[:, indint]
        int_val = val[:, indint] + ac_diff * indratio
        int_values.append(int_val)
    return int_values


def get_atmospheric_rho(targets: typing.Iterable[str],
                        atm_corr_data: AtmCorrData) -> NumValuesDict:
    """Compute atmospheric correction from LUT"""
    sza_lut = atm_corr_data['SZA_LUT']
    oza_lut = atm_corr_data['OZA_LUT']
    delta_lut = atm_corr_data['DELTA_LUT']
    rho_lut = numpy.array(atm_corr_data['RHO_LUT'])

    oza_resol = 5
    # sza_resol = 5
    delta_resol = 10
    ntot_delta = 360 / delta_resol + 1
    ntot_oza = 90 / oza_resol + 1

    sza = atm_corr_data['SZA'].astype('float32')
    saa = atm_corr_data['SAA'].astype('float32')
    oza = atm_corr_data['OZA'].astype('float32')
    oaa = atm_corr_data['OAA'].astype('float32')

    bands_index = atm_corr_data['bands_index']

    # Check the size of the LUT
    lut_size = len(rho_lut)
    lut_size_factor = ntot_oza * ntot_delta
    if 0 != (lut_size % lut_size_factor):
        raise IncompatibleLUTSize(lut_size, lut_size_factor)

    # Interpolate atmospheric correction according to angles
    delta, trunc_to_0 = mod360(saa - oaa, 9)
    _points = numpy.array((sza.flatten(), oza.flatten(),
                           delta.flatten())).transpose()
    sza_lut = numpy.unique(numpy.array(sza_lut, dtype='float32'))
    oza_lut = numpy.unique(numpy.array(oza_lut, dtype='float32'))
    delta_lut = numpy.unique(numpy.array(delta_lut, dtype='float32'))
    lut_points = (sza_lut, oza_lut, delta_lut)

    result = {}
    interpolator = scipy.interpolate.RegularGridInterpolator
    lut_shape = (sza_lut.size, oza_lut.size, delta_lut.size)
    for var_id in targets:
        band_index = bands_index[var_id]
        lut_values = numpy.array(rho_lut[:, band_index],
                                 dtype='float32').reshape(lut_shape)
        lut_func = interpolator(lut_points, lut_values, method='linear',
                                bounds_error=False, fill_value=None)
        band_rho = lut_func(_points).reshape(sza.shape).astype('float32')
        result[var_id] = band_rho
    return result


def studpdf(z: numpy.typing.NDArray,
            wind: float,
            n: float
            ) -> numpy.typing.NDArray:
    """
    """
    # From Cox & Munk [1954]: from optical observations of sun glitter, when
    # all wave lengths are observed:
    #
    # mss <up/down wind,clean surface> = 0.000 + 0.00316 * U10 +/-0.004
    # mss <cross wind,clean surface>   = 0.003 + 0.00192 * U10 +/-0.002
    # mss <CM,clean surface> = mss<up,clean surface> + mss<cx,clean surface>
    #                        = 0.003 + 0.00512 * U10 +/-0.004
    mss = 0.003 + 5.12e-3 * wind

    # Underflow may happen in numpy internal computations when dividing
    # 2z^2 by (mss.(n - 1)), so scale the numerator up by a large number before
    # dividing and scale it down afterwards
    unscaled_operand = 2. * z ** 2.
    scaled_operand = UFLOW_FIX * unscaled_operand
    scaled_operand = scaled_operand / (mss * (n - 1.))
    unscaled_operand = scaled_operand / UFLOW_FIX

    p_denom = (1. + unscaled_operand) ** (n / 2. + 1)
    p: numpy.ndarray = n / (n - 1.) / (numpy.pi * mss) / p_denom  # type:ignore

    return p


def compute_tilt(sza: numpy.typing.ArrayLike,
                 saa: numpy.typing.ArrayLike,
                 oza: numpy.typing.ArrayLike,
                 oaa: numpy.typing.ArrayLike
                 ) -> typing.Tuple[numpy.typing.ArrayLike,
                                   numpy.typing.ArrayLike]:
    """"""
    cos_sza = numpy.cos(numpy.deg2rad(sza))
    sin_sza = numpy.sin(numpy.deg2rad(sza))
    cos_saa = numpy.cos(numpy.deg2rad(saa))
    sin_saa = numpy.sin(numpy.deg2rad(saa))
    cos_oza = numpy.cos(numpy.deg2rad(oza))
    sin_oza = numpy.sin(numpy.deg2rad(oza))
    cos_oaa = numpy.cos(numpy.deg2rad(oaa))
    sin_oaa = numpy.sin(numpy.deg2rad(oaa))

    # Promoting to float64 is required to avoid underflow errors
    tie_zx_num = (sin_sza * cos_saa + sin_oza * cos_oaa).astype(numpy.float64)
    tie_zx_denom = (cos_sza + cos_oza).astype(numpy.float64)
    tie_zx = -1 * tie_zx_num / tie_zx_denom

    # Promoting to float64 is required to avoid underflow errors
    tie_zy_num = (sin_sza * sin_saa + sin_oza * sin_oaa).astype(numpy.float64)
    tie_zy_denom = (cos_sza + cos_oza).astype(numpy.float64)
    tie_zy = -1 * tie_zy_num / tie_zy_denom

    return (tie_zx, tie_zy)


def brightness_contrast(input_opts: InputOptions,
                        output_opts: OutputOptions,
                        granule: Granule,
                        *args: typing.Iterable[typing.Any],
                        targets: typing.Iterable[str],
                        atm_corr_data: AtmCorrData) -> FormatterResult:
    """"""
    sza = atm_corr_data['SZA'].astype('float32')
    saa = atm_corr_data['SAA'].astype('float32')
    oza = atm_corr_data['OZA'].astype('float32')
    oaa = atm_corr_data['OAA'].astype('float32')

    # Extra-terrestrial solar irradiance
    solar_flux = atm_corr_data['solar_flux']

    ac_subsampling = atm_corr_data['ac_subsampling']
    ncell = atm_corr_data['ncell']

    tie_zx, tie_zy = compute_tilt(sza, saa, oza, oaa)

    tie_nrow, tie_ncell = numpy.shape(sza)
    ac_index = numpy.arange(0, ncell, dtype='float32')

    sin_oaa = numpy.sin(numpy.deg2rad(oaa))
    _oza = oza * numpy.sign(sin_oaa)
    _values_list = [tie_zx, tie_zy, sza, _oza]
    zx, zy, interp_sza, interp_oza = interp_tie_ac(_values_list,
                                                   ac_subsampling, ac_index)
    zx_sqr: numpy.typing.ArrayLike = zx ** 2.
    zy_sqr: numpy.typing.ArrayLike = zy ** 2.
    z = numpy.sqrt(zx_sqr + zy_sqr)
    del zx, zy
    interp_oza = numpy.abs(interp_oza)

    # Magic numbers, yeah!
    offset = 0.008263
    k = 0.859117
    w = 6.766508
    n = 4.878770

    # More magic numbers in formula, yeah!
    _stu_pdf = studpdf(z, w, n)
    cos_interp_sza = numpy.cos(numpy.deg2rad(interp_sza))
    cos_interp_oza = numpy.cos(numpy.deg2rad(interp_oza))
    rho_back_numerator = k * _stu_pdf * numpy.pi * 0.022 * (1. + z ** 2.) ** 2
    rho_back_denominator = 4. * cos_interp_sza * cos_interp_oza

    # Underflow may happen in numpy internal computations when dividing
    # rho_back_numerator by rho_back_denominator, so scale the numerator up by
    # a large number before dividing and scale it down afterwards
    scaled_rho_back_numerator = UFLOW_FIX * rho_back_numerator
    scaled_rho_back_nooffset = scaled_rho_back_numerator / rho_back_denominator
    rho_back = offset + scaled_rho_back_nooffset / UFLOW_FIX

    rho_atm = get_atmospheric_rho(targets, atm_corr_data)

    for var_id in targets:
        _band_rho_atm = rho_atm[var_id]
        band_rho_atm = interp_tie_ac([_band_rho_atm], ac_subsampling,
                                     ac_index)[0]
        band_rho_back = rho_back + band_rho_atm

        band = granule.vars[var_id]['array']

        # 10**(-3) variation for solar flux is neglectable so we consider the
        # mean
        es = numpy.mean(solar_flux[var_id])

        rho_band = band * numpy.pi

        # Underflow may happen in numpy internal computations when dividing
        # rho_band by cos_interp_sza and then by es, so scale the numerator up
        # by a large number before dividing and scale it down afterwards
        scaled_rho_band = UFLOW_FIX * rho_band / cos_interp_sza
        scaled_rho_band = scaled_rho_band / es
        result = scaled_rho_band / band_rho_back / UFLOW_FIX

        var_id_bc = f'{var_id}_bc'
        granule.vars[var_id_bc]['array'] = result

    return input_opts, output_opts, granule



# vim: ts=4:sts=4:sw=4
#
# @date 2021-01-13
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import os
import numpy
import numpy.typing
import typing
import netCDF4
import logging
import idf_converter.lib
import idf_converter.readers.netcdf_grid_yx
from idf_converter.lib.types import InputOptions, OutputOptions, ReaderResult

logger = logging.getLogger()

DATA_MODEL = idf_converter.readers.netcdf_grid_yx.DATA_MODEL


def help() -> typing.Tuple[str, str]:
    """Describe options supported by this reader.

    Returns
    -------
    str
        Description of the supported options
    """
    in_msg, out_msg = idf_converter.readers.netcdf_grid_yx.help()
    return (in_msg, out_msg)


def mask_where_no_ice(var_dicts: typing.Dict[str, typing.Any],
                      var_id: str
                      ) -> numpy.typing.NDArray:
    """Build mask for ice concentration.

    Parameters
    ----------
    var_dicts: dict
        Dictionary containing the values for each variable

    Returns
    -------
    numpy.ndarray
        Array with cells set to True where values should be masked , i.e. NaN
        or no ice.
    """
    var_dict = var_dicts[var_id]
    values = var_dict['array']
    values._sharedmask = False
    nan_mask = (numpy.isnan(values))
    values[numpy.where(nan_mask)] = -1.0
    zero_mask = (values <= 0.0)
    _mask = (nan_mask | zero_mask)
    result: numpy.typing.NDArray = numpy.ma.getdata(_mask)
    return result


def read_data(input_opts: InputOptions,
              output_opts: OutputOptions
              ) -> typing.Iterator[ReaderResult]:
    """Read input file, extract data and metadata, store them in a Granule
    object then prepare formatting instructions to finalize the conversion to
    IDF format.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule, list)
        A tuple which contains four elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the extracted information
          has been stored

        - a :obj:list of :obj:dict describing the formatting operations that
          the converter must perform before serializing the result in IDF
          format
    """
    _input_path = input_opts.get('path', None)
    if _input_path is None:
        raise idf_converter.readers.netcdf_grid_yx.InputPathMissing()
    input_path = os.path.normpath(_input_path)

    f_handler = netCDF4.Dataset(input_path, 'r')
    platform = f_handler.platform_name
    sensor = f_handler.instrument_type
    proj4_str = f_handler.variables['Polar_Stereographic_Grid'].proj4_string
    time_bounds = f_handler.variables['time_bnds'][0, :]
    time_bounds_units = f_handler.variables['time_bnds'].units

    # Older versions of the product had a confidence_level variable to use as
    # a quality flag
    uses_old_format = 'confidence_level' in f_handler.variables.keys()
    f_handler.close()

    input_opts['time_variable'] = 'time'
    input_opts['x_variable'] = 'xc'
    input_opts['y_variable'] = 'yc'
    input_opts['spatial_resolution'] = 10000
    input_opts['projection'] = proj4_str
    input_opts['variables'] = 'ice_conc'

    ice_conc_options = 'valid_min:0.0,valid_max:100.0'
    if uses_old_format is True:
        # 0 -> not processed, no input data
        # 1 -> computation failed
        # 2 -> processed but to be used with care
        # 3 -> nominal processing, acceptable quality
        # 4 -> nominal processing, good quality
        # 5 -> nominal processing, excellent quality
        #
        # Only keep nominal processing, i.e. mask where confidence_level < 3
        confidence_mask = 'flag_variable:confidence_level,flag_min:3'
        ice_conc_options = f'{ice_conc_options},{confidence_mask}'
    input_opts['override(ice_conc)'] = ice_conc_options

    dt = netCDF4.num2date(time_bounds, time_bounds_units)
    tc_start = f'time_coverage_start:"{dt[0]:%Y-%m-%dT%H:%M:%SZ}"'
    tc_end = f'time_coverage_end:"{dt[1]:%Y-%m-%dT%H:%M:%SZ}"'
    if 'global_overrides' not in input_opts.keys():
        input_opts['global_overrides'] = f'{tc_start},{tc_end}'
    else:
        tmp = input_opts['global_overrides']
        if 'time_coverage_start' not in tmp:
            tmp = f'{tmp},{tc_start}'
        if 'time_coverage_end' not in tmp:
            tmp = f'{tmp},{tc_end}'
        input_opts['global_overrides'] = tmp

    grid_yx = idf_converter.readers.netcdf_grid_yx
    result = grid_yx.read_data(input_opts, output_opts)
    input_opts, output_opts, granule, transforms = next(result)

    # Convert x/y into meters to match the default unit of the projection
    granule.vars['x']['array'] *= 1000.0
    granule.vars['y']['array'] *= 1000.0

    granule.meta['platform'] = platform
    granule.meta['sensor'] = sensor

    # add mask where there is no ice
    mask_methods = {'ice_conc': mask_where_no_ice}
    transforms.insert(0, ('mask_methods', {'targets': ('ice_conc',),
                                           'methods': mask_methods}))
    yield (input_opts, output_opts, granule, transforms)



# vim: ts=4:sts=4:sw=4
#
# @date 2022-02-24
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import os
import gzip
import json
import numpy
import typing
import logging
import datetime
import operator
import functools
import numpy.typing
import idf_converter.lib
from pkg_resources import resource_filename
from idf_converter.lib.types import InputOptions, OutputOptions, ReaderResult
from idf_converter.readers.remss_l3_wind_netcdf import parse_wind_vars
from idf_converter.readers.remss_l3_wind_netcdf import split_swaths
from idf_converter.readers.remss_l3_wind_netcdf import create_granule
from idf_converter.readers.remss_l3_wind_netcdf import get_wind_transforms

logger = logging.getLogger(__name__)

DATA_MODEL = idf_converter.readers.netcdf_grid_latlon.DATA_MODEL
MutableIndexSlices = typing.List[typing.Union[int, slice]]
IndexSlices = typing.Sequence[typing.Union[int, slice]]
SPEED_VARS = ('WSPD', 'WSPD_LF', 'WSPD_MF', 'WSPD_AW')

"""
Filename prefixes according to REMSS naming convention:

f08: SSMI
f10: SSMI
f11: SSMI
f12: TMI
f13: SSMI
f14: SSMI
f15: SSMI
f16: SSMI
f17: SSMI
f18: SSMI
f32: AMSR-E
f33: AMSR-J
f34: AMSR-2
f35: GMI

qscat: QuikScat
ascat: MetOp-A
ascatb: MetOp-B
ascatc: MetOp-C
"": SeaWinds
wsat: WindSat
"""


class InputPathMissing(Exception):
    """Error raised when the "path" input option has not been specified by the
    user."""
    pass


class MissionMissing(Exception):
    """Error raised when the 'mission' input option has not been specified by
    the user."""
    pass


class TimeUnitsNotSupported(Exception):
    """Error raised when the units of the time variable are not support by this
    reader."""
    def __init__(self, time_units: str) -> None:
        self.time_units = time_units


class InvalidWindVariable(ValueError):
    """Error raised when the value passed to the wind_variables option does not
    match the expected pattern."""
    def __init__(self, value: str) -> None:
        self.wind_variables_str = value


def help() -> typing.Tuple[str, str]:
    """Describe options supported by this reader.

    Returns
    -------
    str
        Description of the supported options
    """
    inp = ('    path\tPath of the input file',)
    out = ('    split-swaths\tSet to yes|true|1 if the converter should create'
           ' one output file for each swath (simulating a gridded L2) instead'
           ' of the original data structure (one file containing all swaths)',
           )
    return ('\n'.join(inp), '\t'.join(out))


def get_day_from_filename(input_path: str
                          ) -> datetime.datetime:
    """"""
    file_name = os.path.basename(input_path)
    name_items = file_name.split('_')
    date_str = name_items[-1]
    if date_str.startswith('v'):
        # For ASCAT data the last item contains the version number, not the
        # date.
        date_str = name_items[-2]

    year = int(date_str[0:4])
    month = int(date_str[4:6])
    day = int(date_str[6:8])

    dt = datetime.datetime(year, month, day, 0, 0, 0)
    return dt


def read_data(input_opts: InputOptions,
              output_opts: OutputOptions
              ) -> typing.Iterator[ReaderResult]:
    """Read input file, extract data and metadata, store them in a Granule
    object then prepare formatting instructions to finalize the conversion to
    IDF format.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule, list)
        A tuple which contains four elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the extracted information
          has been stored

        - a :obj:list of :obj:dict describing the formatting operations that
          the converter must perform before serializing the result in IDF
          format
    """
    _input_path = input_opts.get('path', None)
    if _input_path is None:
        raise InputPathMissing()
    input_path = os.path.normpath(_input_path)

    mission = input_opts.get('mission', None)
    if mission is None:
        raise MissionMissing()

    variables = input_opts.get('variables', None)
    original_var_ids = [_.strip() for _ in variables.split(',')
                        if 0 < len(_.strip())]

    required_var_ids, wind_vars = parse_wind_vars(input_opts)

    day = get_day_from_filename(input_path)

    var_ids = original_var_ids + required_var_ids
    var_ids = list(set(var_ids))
    if 0 >= len(var_ids):
        logger.warning('No variable to extract')
        raise idf_converter.lib.EarlyExit()

    with gzip.open(input_path, 'rb') as f:
        binary_data = f.read()

    offset = 0
    all_values = {}
    mappings_path = resource_filename('idf_converter',
                                      'share/remss_bytemaps.json')
    with open(mappings_path, 'r') as f:
        mappings = json.load(f)
    struct = mappings.get(mission, None)
    shape = struct['shape']
    for var_id, var_cfg in struct['vars'].items():
        add_offset = var_cfg['offset']
        scale_factor = var_cfg['scale_factor']
        valid_min = var_cfg['valid_min']
        valid_max = var_cfg['valid_max']

        count = functools.reduce(operator.mul, shape)
        packed_data = numpy.frombuffer(binary_data, dtype='ubyte',
                                       offset=offset,
                                       count=count).reshape(shape)
        offset = offset + count

        packed_data = numpy.ma.masked_where(packed_data > 250, packed_data)
        has_rain_flag = (mission in ('ascat', 'quikscat', 'seaswinds'))
        if has_rain_flag and ('RAIN' == var_id):
            suffix = var_id.replace('RAIN_', '')

            scat_rain = ((packed_data & 0x01) > 0)
            var_name = f'SCATFLAG_{suffix}'
            all_values[var_name] = {'array': scat_rain,
                                    'units': '1',
                                    'valid_min': 0,
                                    'valid_max': 1,
                                    'long_name': 'Scatterometer rain flag',
                                    'options': {}}

            adjacent_rain = ((packed_data & 0x02) > 0)
            var_name = f'RADFLAG_{suffix}'
            all_values[var_name] = {'array': adjacent_rain,
                                    'units': '1',
                                    'valid_min': 0,
                                    'valid_max': 1,
                                    'long_name': 'Collocated radiometer flag',
                                    'options': {}}

            packed_data = (packed_data & 0b11111100) >> 2

        data = add_offset + scale_factor * packed_data
        data = numpy.ma.masked_where((data < valid_min) | (data > valid_max),
                                     data)
        all_values[var_id] = {'array': data,
                              'units': var_cfg['units'],
                              'valid_min': valid_min,
                              'valid_max': valid_max,
                              'long_name': var_cfg['long_name'],
                              'options': {}}

    minutes_asc = all_values['TIME_asc']['array']
    minutes_desc = all_values['TIME_desc']['array']

    lat0 = struct['lat0']
    dlat = struct['dlat']
    lon0 = struct['lon0']
    dlon = struct['dlon']
    lat = numpy.array([lat0 + i * dlat for i in range(0, shape[0])])
    lon = numpy.array([lon0 + i * dlon for i in range(0, shape[1])])
    spatial_resolution = min([abs(dlat), abs(dlon)]) * 111000.

    use_swaths_splitting = output_opts.get('split-swaths', 'n')
    if use_swaths_splitting.lower() in ('1', 'y', 'yes'):
        asc_values = {}
        asc_units = {}
        for var_id, var_values in all_values.items():
            if var_id.endswith('_asc'):
                simple_var_id, _ = var_id.rsplit('_', 1)
                asc_values[simple_var_id] = var_values['array']
                asc_units[simple_var_id] = var_values['units']
        yield from split_swaths(input_opts, output_opts, original_var_ids,
                                day, minutes_asc, lat, lon, '_asc',
                                asc_values, asc_units, wind_vars)

        desc_values = {}
        desc_units = {}
        for var_id, var_values in all_values.items():
            if var_id.endswith('_desc'):
                simple_var_id, _ = var_id.rsplit('_', 1)
                desc_values[simple_var_id] = var_values['array']
                desc_units[simple_var_id] = var_values['units']
        yield from split_swaths(input_opts, output_opts, original_var_ids,
                                day, minutes_desc, lat, lon, '_desc',
                                desc_values, desc_units, wind_vars)
    else:
        # Time coverage
        minutes = numpy.vstack([minutes_asc, minutes_desc])
        start_minutes = int(minutes.min())
        start_dt = day + datetime.timedelta(seconds=start_minutes * 60)

        stop_minutes = int(minutes.max())
        stop_dt = day + datetime.timedelta(seconds=stop_minutes * 60)

        values = {}
        units = {}
        all_var_ids = list(all_values.keys())
        for var_id in all_var_ids:
            simple_var_id, _ = var_id.rsplit('_', 1)
            if simple_var_id in var_ids:
                values[var_id] = all_values[var_id]['array']
                units[var_id] = all_values[var_id]['units']

        granule = create_granule(input_opts, output_opts, '',
                                 start_dt, stop_dt, spatial_resolution,
                                 lat, lon, values, units)

        output_opts['__export'] = list(values.keys())

        transforms = get_wind_transforms(output_opts, wind_vars,
                                         original_var_ids, True)

        yield input_opts, output_opts, granule, transforms



# vim: ts=4:sts=4:sw=4
#
# @date 2021-12-28
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import os
import numpy
import pyproj
import typing
import netCDF4
import logging
import datetime
import idf_converter.lib
from idf_converter.lib.types import InputOptions, OutputOptions
from idf_converter.lib.types import ReaderResult, TransformsList

logger = logging.getLogger(__name__)

DATA_MODEL = 'time_dependent'
VARS_IN_KNOTS = ('wind',)
VARS_IN_MB = ('pres',)
VARS_IN_NM = ('r34_ne', 'r34_se', 'r34_sw', 'r34_nw', 'r50_ne', 'r50_se',
              'r50_sw', 'r50_nw', 'r64_ne', 'r64_se', 'r64_sw', 'r64_nw',
              'rmw')


class InputPathMissing(ValueError):
    """Error raised when the input options have no "path" entry"""
    pass


class NotEnoughData(Exception):
    """Error raised when the input data has less than 3 values, which is the
    minimal values count supported by this reader."""
    def __init__(self) -> None:
        """"""
        msg = 'This reader requires at least three along-track values'
        super(NotEnoughData, self).__init__(msg)


class StormNameMissing(Exception):
    """Error raised when the storm name variable contains only empty values."""
    def __init__(self, storm_id: str) -> None:
        """"""
        self.storm_id = storm_id


class StormNameConflict(Exception):
    """Error raised when the storm name variable contains different, non-empty
    values."""
    def __init__(self, storm_id: str) -> None:
        """"""
        self.storm_id = storm_id


class StormATCFIDMissing(Exception):
    """Error raised when the storm ATCF identifier variable contains only empty
    values."""
    def __init__(self, storm_id: str) -> None:
        """"""
        self.storm_id = storm_id


class StormATCFIDConflict(Exception):
    """Error raised when the storm ATCF identifier variable contains different,
    non-empty values."""
    def __init__(self, storm_id: str) -> None:
        """"""
        self.storm_id = storm_id


class InvalidMaxTranslationSpeed(ValueError):
    """Error raised when the value passed for the max_translation_speed input
    option cannot be parsed as a float."""
    def __init__(self, value: str) -> None:
        """"""
        self.invalid_value = value


def help() -> typing.Tuple[str, str]:
    """Describe options supported by this reader.

    Returns
    -------
    str
        Description of the supported options
    """
    inp = ('    path\tPath of the input file',
           '    max_translation_mps\tThreshold for the storm translation '
           'speed: observations for which the translation speed exceeds the '
           'threshold will be discarded. Expressed in meter per second, '
           'default value: 83.3333 (i.e. 300km/h)')
    out = ('',)
    return ('\n'.join(inp), '\n'.join(out))


def read_data(input_opts: InputOptions,
              output_opts: OutputOptions
              ) -> typing.Iterator[ReaderResult]:
    """Read input file, extract data and metadata, store them in a Granule
    object then prepare formatting instructions to finalize the conversion to
    IDF format.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule, list)
        A tuple which contains four elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the extracted information
          has been stored

        - a :obj:list of :obj:dict describing the formatting operations that
          the converter must perform before serializing the result in IDF
          format
    """
    _convertutc = datetime.datetime.utcfromtimestamp

    variables = {'wind': {'name': 'wind'},
                 'pres': {'name': 'pressure'},
                 'r34_ne': {'name': 'r34_ne'},
                 'r34_se': {'name': 'r34_se'},
                 'r34_sw': {'name': 'r34_sw'},
                 'r34_nw': {'name': 'r34_nw'},
                 'r50_ne': {'name': 'r50_ne'},
                 'r50_se': {'name': 'r50_se'},
                 'r50_sw': {'name': 'r50_sw'},
                 'r50_nw': {'name': 'r50_nw'},
                 'r64_ne': {'name': 'r64_ne'},
                 'r64_se': {'name': 'r64_se'},
                 'r64_sw': {'name': 'r64_sw'},
                 'r64_nw': {'name': 'r64_nw'},
                 'rmw': {'name': 'rmax'},
                 }

    idf_version = output_opts.get('idf_version', '1.0')
    granule = idf_converter.lib.create_granule(idf_version, DATA_MODEL)

    _input_path = input_opts.get('path', None)
    if _input_path is None:
        raise InputPathMissing()
    input_path = os.path.normpath(_input_path)

    max_mps_str = input_opts.get('max_translation_mps', str(300000 / 3600))
    try:
        max_mps = float(max_mps_str)
    except ValueError:
        raise InvalidMaxTranslationSpeed(max_mps_str)

    granule.vars = variables
    channels = list(variables.keys())

    spatial_res_meters = 1.e07

    f_handler = netCDF4.Dataset(input_path, 'r')
    _time = idf_converter.lib.extract_variable_values(f_handler, 'time')
    _lon = idf_converter.lib.extract_variable_values(f_handler, 'lon')
    _lat = idf_converter.lib.extract_variable_values(f_handler, 'lat')
    _atcf_id = netCDF4.chartostring(f_handler.variables['atcf_id'][:])
    _sid = netCDF4.chartostring(f_handler.variables['sid'][:])
    _name = netCDF4.chartostring(f_handler.variables['name'][:])

    _lon_mask = numpy.ma.getmaskarray(_lon)
    if _lon_mask.any():
        logger.warning(f'Masked longitudes in {input_path}')
    _lon = numpy.ma.getdata(_lon)

    _lat_mask = numpy.ma.getmaskarray(_lat)
    if _lat_mask.any():
        logger.warning(f'Masked latitudes in {input_path}')
    _lat = numpy.ma.getdata(_lat)

    _time_mask = numpy.ma.getmaskarray(_time)
    if _time_mask.any():
        logger.warning(f'Masked times in {input_path}')
    _time = numpy.ma.getdata(_time)

    # Apply time offset to transform time into an EPOCH timestamp
    ref_time = datetime.datetime(2010, 1, 1, 12, 0, 0)
    epoch = datetime.datetime(1970, 1, 1)
    time_offset = (ref_time - epoch).total_seconds()
    shifted_time = numpy.int_(time_offset) + _time * 60
    _time = shifted_time.astype(numpy.double)

    # Extract data variables
    _data = {}
    for var_id in granule.vars.keys():
        idf_converter.lib.extract_variable_attributes(f_handler, var_id,
                                                      granule)
        band = idf_converter.lib.extract_variable_values(f_handler, var_id)
        _data[var_id] = band
    f_handler.close()

    # Group observations based on storm identifiers
    sorted_ind = _sid.argsort(0)
    sorted_sid = _sid[sorted_ind]
    unique_sid, sid_slice_start = numpy.unique(sorted_sid, return_index=True)
    sid_ind = numpy.split(sorted_ind, sid_slice_start[1:])

    geod = pyproj.Geod(ellps='WGS84')
    for i, storm_id in enumerate(unique_sid):
        # Reorder indices associated to this storm by ascending time
        storm_ind = sid_ind[i]
        unsorted_storm_time = _time[storm_ind]
        time_sorted_storm_ind = storm_ind[unsorted_storm_time.argsort(0)]

        storm_time = _time[time_sorted_storm_ind]
        storm_lat = _lat[time_sorted_storm_ind]
        storm_lon = _lon[time_sorted_storm_ind]

        # Longitude continuity fix (required for distance computation later on)
        lon0 = numpy.mean(storm_lon)
        storm_lon = lon0 + numpy.mod(storm_lon - lon0, 360)
        storm_lon[numpy.where(storm_lon > lon0 + 180)] -= 360
        storm_lon[numpy.where(storm_lon < lon0 - 180)] += 360

        storm_data = {}
        for var_id in variables:
            storm_data[var_id] = _data[var_id][time_sorted_storm_ind]

        # Remove outliers (translation speed above 150km/h)
        dists = geod.line_lengths(storm_lon, storm_lat)
        storm_dtime = storm_time[1:] - storm_time[:-1]
        outliers_ind = numpy.where(dists > storm_dtime * max_mps)
        while 0 < len(outliers_ind[0]):
            del_ind = outliers_ind[0][0] + 1
            logger.warning(f'Removing outlier for {storm_id}: '
                           f'at {storm_lon[del_ind]}, {storm_lat[del_ind]}')

            # remove outlier from coordinates and variables
            storm_time = numpy.delete(storm_time, del_ind)
            storm_lon = numpy.delete(storm_lon, del_ind)
            storm_lat = numpy.delete(storm_lat, del_ind)
            for var_id in variables:
                storm_data[var_id] = numpy.delete(storm_data[var_id], del_ind)

            # detect remaining outliers
            dists = geod.line_lengths(storm_lon, storm_lat)
            storm_dtime = storm_time[1:] - storm_time[:-1]
            outliers_ind = numpy.where(dists > storm_dtime * max_mps)

        # extract a single name for the storm
        _storm_name = _name[time_sorted_storm_ind]
        non_empty_ind = numpy.where(_storm_name != '')
        _storm_name = numpy.unique(_storm_name[non_empty_ind])
        if 0 >= _storm_name.size:
            raise StormNameMissing(storm_id)
        elif 1 < _storm_name.size:
            raise StormNameConflict(storm_id)
        storm_name = _storm_name[0]

        # extract a single ATCF identifier for the storm
        _storm_atcf_id = _atcf_id[time_sorted_storm_ind]
        non_empty_ind = numpy.where(_storm_atcf_id != '')
        _storm_atcf_id = numpy.unique(_storm_atcf_id[non_empty_ind])
        storm_atcf_id = None
        if 1 < _storm_atcf_id.size:
            raise StormATCFIDConflict(storm_id)
        elif 1 == _storm_atcf_id.size:
            storm_atcf_id = _storm_atcf_id[0]

        storm_granule = idf_converter.lib.create_granule(idf_version,
                                                         DATA_MODEL)

        for var_id in variables:
            storm_granule.vars[var_id] = {'options': {}}

            # Copy attributes from the "main" granule object
            for attr_name, attr_value in granule.vars[var_id].items():
                if 'array' == attr_name:
                    continue
                storm_granule.vars[var_id][attr_name] = attr_value

            band = storm_data[var_id]
            vmin = storm_granule.vars[var_id]['valid_min']
            vmax = storm_granule.vars[var_id]['valid_max']

            # 1 knot = 1 international nautical mile per hour
            if var_id in VARS_IN_KNOTS:
                band = band * numpy.float64(1852.0 / 3600.0)
                storm_granule.vars[var_id]['units'] = 'm.s-1'
                storm_granule.vars[var_id]['valid_min'] = vmin * 1852 / 3600
                storm_granule.vars[var_id]['valid_max'] = vmax * 1852 / 3600

            # 1 international nautical mile = 1852 meters
            if var_id in VARS_IN_NM:
                band = band * 1852
                storm_granule.vars[var_id]['units'] = 'm'
                storm_granule.vars[var_id]['valid_min'] = vmin * 1852
                storm_granule.vars[var_id]['valid_max'] = vmax * 1852

            # 1 millibar = 100 Pa
            if var_id in VARS_IN_MB:
                band = band * 100
                storm_granule.vars[var_id]['units'] = 'Pa'
                storm_granule.vars[var_id]['valid_min'] = vmin * 100
                storm_granule.vars[var_id]['valid_max'] = vmax * 100

            storm_granule.vars[var_id]['array'] = band

        # Add coordinates variables
        storm_granule.vars['lat'] = {'units': 'degrees north',
                                     'array': storm_lat,
                                     'datatype': _lat.dtype,
                                     'options': {}}
        storm_granule.vars['lon'] = {'units': 'degrees east',
                                     'array': storm_lon,
                                     'datatype': _lon.dtype,
                                     'options': {}}

        out_time_units = 'seconds since 1970-01-01T00:00:00.000Z'
        storm_granule.vars['time'] = {'units': out_time_units,
                                      'array': storm_time,
                                      'datatype': numpy.double,
                                      'options': {}}

        # Set Global parameters
        granule_name = os.path.splitext(os.path.basename(input_path))[0]
        storm_granule.meta['idf_subsampling_factor'] = 0
        storm_granule.meta['idf_spatial_resolution'] = spatial_res_meters
        storm_granule.meta['idf_spatial_resolution_units'] = 'm'

        if storm_atcf_id is not None:
            storm_granule.meta['atcf_id'] = storm_atcf_id
        storm_granule.meta['storm_name'] = storm_name
        storm_granule.meta['storm_id'] = storm_id

        start_dt = _convertutc(storm_time[0])
        stop_dt = _convertutc(storm_time[-1])
        if storm_name not in ('', 'NOT NAMED', 'NOT_NAMED'):
            _granule_name = f'{storm_id}_{storm_name}_{granule_name}'
        else:
            _granule_name = f'{storm_id}_{granule_name}'
        storm_granule.meta['idf_granule_id'] = f'{_granule_name}'
        storm_granule.meta['time_coverage_start'] = start_dt
        storm_granule.meta['time_coverage_end'] = stop_dt

        transforms: TransformsList = []

        transforms.append(('remove_extra_lon_degrees', {'lon_name': 'lon'}))

        output_opts['__export'] = channels

        # data_model, dims, vars, attrs, formatter_jobs
        yield (input_opts, output_opts, storm_granule, transforms)



# vim: ts=4:sts=4:sw=4
#
# @date 2019-10-10
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import os
import numpy
import numpy.typing
import typing
import pyproj
import netCDF4
import logging
import datetime
import idf_converter.lib
import idf_converter.readers.netcdf_grid_yx
from idf_converter.lib.types import InputOptions, OutputOptions, ReaderResult

logger = logging.getLogger()

DATA_MODEL = idf_converter.readers.netcdf_grid_yx.DATA_MODEL


class FilenamePatternNotSupported(ValueError):
    """Error raised when the name of the input file does nt match any known
    pattern, preventing the detecting of the time coverage, platform and
    sensor."""
    def __init__(self, file_name: str) -> None:
        """"""
        self.file_name = file_name


def help() -> typing.Tuple[str, str]:
    """Describe options supported by this reader.

    Returns
    -------
    str
        Description of the supported options
    """
    in_msg, out_msg = idf_converter.readers.netcdf_grid_yx.help()
    return (in_msg, out_msg)


def mask_where_no_ice(var_dicts: typing.Dict[str, typing.Any],
                      var_id: str) -> numpy.typing.NDArray:
    """Build mask for ice concentration.

    Parameters
    ----------
    var_dict: dict
        Dictionary describing the sea ice concentration variable

    Returns
    -------
    numpy.ndarray
        Array with cells set to True where values should be masked , i.e. NaN
        or aberrant values (absolute values above 50m).
    """
    var_dict = var_dicts[var_id]
    values = var_dict['array']
    values._sharedmask = False
    nan_mask = (numpy.isnan(values))
    values[numpy.where(nan_mask)] = -1.0
    zero_mask = (values <= 0.0)
    _mask = (nan_mask | zero_mask)
    mask: numpy.typing.NDArray = numpy.ma.getdata(_mask)
    return mask


def read_data(input_opts: InputOptions,
              output_opts: OutputOptions
              ) -> typing.Iterator[ReaderResult]:
    """Read input file, extract data and metadata, store them in a Granule
    object then prepare formatting instructions to finalize the conversion to
    IDF format.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule, list)
        A tuple which contains four elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the extracted information
          has been stored

        - a :obj:list of :obj:dict describing the formatting operations that
          the converter must perform before serializing the result in IDF
          format
    """
    _input_path = input_opts.get('path', None)
    if _input_path is None:
        raise idf_converter.readers.netcdf_grid_yx.InputPathMissing()
    input_path = os.path.normpath(_input_path)

    logger.warning('Guessing granule time coverage from the input file '
                   'name is not a good practice: it is only used here '
                   'because time coverage information is missing from the '
                   'netCDF content.')
    file_name = os.path.basename(input_path)
    granule_name, _ = os.path.splitext(file_name)
    name_elements = granule_name.split('-')
    start_dt = None
    platform = None
    sensor = None
    if 4 == len(name_elements):
        start_dt = datetime.datetime.strptime(name_elements[2], '%Y%m%d')
        platform = 'Aqua'
        sensor = 'AMSR-E'
    elif 5 == len(name_elements):
        start_dt = datetime.datetime.strptime(name_elements[3], '%Y%m%d')
        platform = 'GCOM-W1'
        sensor = 'AMSR-2'
    else:
        raise FilenamePatternNotSupported(file_name)

    stop_dt = start_dt + datetime.timedelta(seconds=86400)

    start_str = f'{start_dt:%Y-%m-%dT%H:%M:%SZ}'
    stop_str = f'{stop_dt:%Y-%m-%dT%H:%M:%SZ}'
    if 'global_overrides' in input_opts:
        input_opts['global_overrides'] += f',time_coverage_start:"{start_str}"'
        input_opts['global_overrides'] += f',time_coverage_end:"{stop_str}"'
    else:
        input_opts['global_overrides'] = f'time_coverage_start:"{start_str}"'
        input_opts['global_overrides'] += f',time_coverage_end:"{stop_str}"'

    if 'x_variable' not in input_opts:
        input_opts['x_variable'] = 'x'
    if 'y_variable' not in input_opts:
        input_opts['y_variable'] = 'y'
    if 'variables' not in input_opts:
        input_opts['variables'] = 'z'
    if 'spatial_resolution' not in input_opts:
        input_opts['spatial_resolution'] = 6250
    if 'projection' not in input_opts:
        f_handler = netCDF4.Dataset(input_path, 'r')
        _crs = f_handler.variables['polar_stereographic']
        crs = pyproj.CRS.from_cf({x: getattr(_crs, x) for x in _crs.ncattrs()})
        input_opts['projection'] = crs.to_wkt()
        f_handler.close()

    grid_yx = idf_converter.readers.netcdf_grid_yx
    for result in grid_yx.read_data(input_opts, output_opts):
        input_opts, output_opts, granule, transforms = result

        granule.meta['platform'] = platform
        granule.meta['sensor'] = sensor

        # add mask where there is no ice
        mask_methods = {'z': mask_where_no_ice}
        transforms.insert(0, ('mask_methods', {'targets': ('z',),
                                               'methods': mask_methods}))
        yield (input_opts, output_opts, granule, transforms)



# vim: ts=4:sts=4:sw=4
#
# @date 2019-09-25
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""This module provides methods to serialize Granules objects into IDF files.
"""

import os
import typing
import logging
import netCDF4
import idf_converter.lib
import idf_converter.lib.downscale
from idf_converter.lib.types import InputOptions, OutputOptions, Granule

logger = logging.getLogger(__name__)


class NCCreateDimError(Exception):
    """Error raised when the netCDF4 module failed to create a dimension on
    the IDF output file."""

    def __init__(self, dim_name: str, dim_size: int) -> None:
        msg = (f'Error while creating netCDF dimension with name={dim_name} '
               f'and size={dim_size}')
        super(NCCreateDimError, self).__init__(msg)


def help() -> typing.Tuple[str, str]:
    """"""
    inp = ('',)
    out = ('    downscale\tRequest production of IDF files at degraded'
           ' resolutions in addition to the full resolution file. Valid'
           ' values are "true", "yes", "1" (defaults to "no"). Please note'
           ' that downscaling is not supported for time-dependent granules.',)
    return ('\n'.join(inp), '\n'.join(out))


def close_and_remove(idf_file: netCDF4.Dataset) -> None:
    """Close and delete an IDF file after an error.

    Parameters
    ----------
    idf_file: netCDF4.Dataset
        NetCDF4 handle for the IDF file to close and delete.
    """
    filepath = idf_file.filepath()
    idf_file.close()
    os.remove(filepath)


def as_idf(original_input_opts: InputOptions,
           original_output_opts: OutputOptions,
           original_granule: Granule) -> None:
    """Create IDF file from granule information.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format
    granule: idf_converter.lib.Granule
        Object representing the granule data and metadata
    """

    fmt = idf_converter.lib.get_idf_formatter(original_granule.idf_version)

    prepared = fmt.preformat(original_input_opts, original_output_opts,
                             original_granule)

    for input_opts, output_opts, granule in prepared:
        # Create dataset
        output_dir, idf_fname = fmt.get_idf_path(input_opts, output_opts,
                                                 granule)
        os.makedirs(output_dir, exist_ok=True)
        output_path = os.path.join(output_dir, idf_fname)
        idf_file = netCDF4.Dataset(output_path, mode='w', format='NETCDF4',
                                   clobber=True)

        # IDF-specific formatting
        formatted = fmt.granule2idf(input_opts, output_opts, granule, idf_file)
        processed_dims, processed_vars, processed_attrs = formatted

        dimensions = granule.dims
        attributes = granule.meta
        variables = granule.vars

        # Add other dimensions
        for dim_name, dim_size in dimensions.items():
            if dim_name in processed_dims:
                continue

            try:
                idf_file.createDimension(dim_name, dim_size)
            except Exception:  # TODO: use more specific exception catching
                close_and_remove(idf_file)
                raise NCCreateDimError(dim_name, dim_size)

            processed_dims.append(dim_name)

        # Add other attributes
        for attrname, attrval in attributes.items():
            if attrname in processed_attrs:
                continue
            setattr(idf_file, attrname, attrval)

        idf_file.sync()

        # Variables
        specs = idf_converter.lib.get_idf_specs(granule.idf_version)
        for var_id in variables.keys():
            if var_id in processed_vars:
                continue
            tmp_res = fmt.write_variable(var_id, granule, idf_file, specs)
            _var, processed_var_attrs = tmp_res
            for var_attr in granule.vars[var_id].keys():
                if var_attr not in processed_var_attrs:
                    logger.warning(f'{var_attr} not processed for {var_id}')
            idf_file.sync()

        logger.debug('Downscaling...')
        _downscale = output_opts.get('downscale', 'no')
        downscale = _downscale.lower() in ('true', 'yes', '1')
        if downscale:
            idf_converter.lib.downscale.downscale(output_path)

        idf_file.close()



# vim: ts=4:sts=4:sw=4
#
# @date 2019-09-25
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import os
import numpy
import numpy.typing
import typing
import netCDF4
import logging
import datetime
import idf_converter.lib
import defusedxml.ElementTree as ET
from idf_converter.lib.types import InputOptions, OutputOptions
from idf_converter.lib.types import ReaderResult

logger = logging.getLogger(__name__)

DATA_MODEL = 'time_dependent'


class InputPathMissing(ValueError):
    """Error raised when the input options have no "path" entry"""
    pass


class FrequencyNotSupported(ValueError):
    """Error raised when the requested frequency is neither 1Hz not 20Hz, the
    only frequencies available in the input files."""
    def __init__(self, frequency: int) -> None:
        """"""
        self.frequency = frequency
        msg = 'This reader only supports 1Hz and 20Hz data'
        super(FrequencyNotSupported, self).__init__(msg)


class TimelinessNotSupported(Exception):
    """Error raised when the input file corresponds to a Non Time Critical
    product and 20Hz data are requested: the reader only supports 20Hz for
    Near Real Time products."""
    def __init__(self, timeliness: str) -> None:
        """"""
        self.timeliness = timeliness


class NotEnoughData(Exception):
    """Error raised when the input data has less than 3 values, which is the
    minimal values count supported by this reader."""
    def __init__(self) -> None:
        """"""
        msg = 'This reader requires at least three along-track values'
        super(NotEnoughData, self).__init__(msg)


def get_timeliness(input_dir: str) -> typing.Optional[str]:
    """ Read metadata in xml file. Return start and track Offset, number of
    lines and elements spatial resolution for Tie Points, 1 km nadir and 1 km
    oblique grid."""
    manifest_path = os.path.join(input_dir, 'xfdumanifest.xml')
    if not os.path.exists(manifest_path):
        return None
    tree = ET.parse(manifest_path)
    root = tree.getroot()
    s3path = '{http://www.esa.int/safe/sentinel/sentinel-3/1.0}'
    for a_child in root.findall("./*/metadataObject/metadataWrap/xmlData/*"):
        for child in a_child:
            if child.tag == f'{s3path}timeliness':
                return str(child.text)
    return None


def help() -> typing.Tuple[str, str]:
    """Describe options supported by this reader.

    Returns
    -------
    str
        Description of the supported options
    """
    inp = ('    path\tPath of the input file',
           '    frequency\tEither 1 or 20 (1Hz by default)',)
    out = ('',)
    return ('\n'.join(inp), '\n'.join(out))


def mask_ssha(var_dicts: typing.Dict[str, typing.Any],
              var_id: str
              ) -> numpy.typing.NDArray:
    """Build mask for SSHA.

    Parameters
    ----------
    var_dict: dict
        Dictionary describing the SSHA variable

    Returns
    -------
    numpy.ndarray
        Array with cells set to True where values should be masked , i.e. NaN
        or aberrant values (absolute values above 50m).
    """
    var_dict = var_dicts[var_id]
    values = var_dict['array']
    values._sharedmask = False
    nan_mask = (numpy.isnan(values))
    values[numpy.where(nan_mask)] = var_dict['_FillValue']  # avoid warning
    out_of_range_mask = (numpy.absolute(values) > 50.)
    _mask = (nan_mask | out_of_range_mask)
    result: numpy.typing.NDArray = numpy.ma.getdata(_mask)
    return result


def mask_swh(var_dicts: typing.Dict[str, typing.Any],
             var_id: str
             ) -> numpy.typing.NDArray:
    """Build mask for SWH.

    Parameters
    ----------
    var_dict: dict
        Dictionary describing the SWH variable

    Returns
    -------
    numpy.ndarray
        Array with cells set to True where values should be masked , i.e. NaN
        or aberrant values (absolute values above 50m).
    """
    var_dict = var_dicts[var_id]
    values = var_dict['array']
    values._sharedmask = False
    nan_mask = (numpy.isnan(values))
    values[numpy.where(nan_mask)] = var_dict['_FillValue']  # avoid warning
    out_of_range_mask = (numpy.absolute(values) > 50.)
    _mask = (nan_mask | out_of_range_mask)
    result: numpy.typing.NDArray = numpy.ma.getdata(_mask)
    return result


def mask_sig0(var_dicts: typing.Dict[str, typing.Any],
              var_id: str
              ) -> numpy.typing.NDArray:
    """Build mask for Sigma0.

    Parameters
    ----------
    var_dict: dict
        Dictionary describing the Sigma0 variable

    Returns
    -------
    numpy.ndarray
        Array with cells set to True where values should be masked , i.e. NaN
        or aberrant values (values above 100).
    """
    var_dict = var_dicts[var_id]
    values = var_dict['array']
    values._sharedmask = False
    nan_mask = (numpy.isnan(values))
    values[numpy.where(nan_mask)] = var_dict['_FillValue']  # avoid warning
    out_of_range_mask = (values > 100.)
    _mask = (nan_mask | out_of_range_mask)
    result: numpy.typing.NDArray = numpy.ma.getdata(_mask)
    return result


def mask_wind(var_dicts: typing.Dict[str, typing.Any],
              var_id: str
              ) -> numpy.typing.NDArray:
    """Build mask for Wind speed.

    Parameters
    ----------
    var_dict: dict
        Dictionary describing the Wind speed variable

    Returns
    -------
    numpy.ndarray
        Array with cells set to True where values should be masked , i.e. NaN
        or aberrant values (absolute values above 200m.s-1).
    """
    var_dict = var_dicts[var_id]
    values = var_dict['array']
    values._sharedmask = False
    nan_mask = (numpy.isnan(values))
    values[numpy.where(nan_mask)] = var_dict['_FillValue']  # avoid warning
    out_of_range_mask = (numpy.absolute(values) > 200.)
    _mask = (nan_mask | out_of_range_mask)
    result: numpy.typing.NDArray = numpy.ma.getdata(_mask)
    return result


def read_data(input_opts: InputOptions,
              output_opts: OutputOptions
              ) -> typing.Iterator[ReaderResult]:
    """Read input file, extract data and metadata, store them in a Granule
    object then prepare formatting instructions to finalize the conversion to
    IDF format.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule, list)
        A tuple which contains four elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the extracted information
          has been stored

        - a :obj:list of :obj:dict describing the formatting operations that
          the converter must perform before serializing the result in IDF
          format
    """
    variables = {'wind': {'name': 'wind_speed',
                          'long_name': 'Wind speed',
                          'units': 'm.s-1',
                          'valid_min': 0.0,
                          'valid_max': 25.4,
                          'options': {}},
                 'ssha': {'name': 'ssha',
                          'long_name': 'Sea surface height anomaly',
                          'units': 'm',
                          'valid_min': -1.0,
                          'valid_max': 1.0,
                          'options': {}},
                 'swh': {'name': 'swh',
                         'long_name': 'Significant wave height',
                         'units': 'm',
                         'valid_min': 0.0,
                         'valid_max': 25.0,
                         'options': {}},
                 'sig0': {'name': 'sigma0',
                          'long_name': 'Sigma0',
                          'units': '',
                          'valid_min': 5.0,
                          'valid_max': 30.0,
                          'options': {}},
                 'tb238': {'name': 'tb_238_01',
                           'long_name': ('23.8Ghz main beam brightness '
                                         'temperature: 1Hz'),
                           'units': 'K',
                           'valid_min': 135.15,
                           'valid_max': 265.15,
                           'options': {}},
                 'tb365': {'name': 'tb_365_01',
                           'long_name': ('36.5Ghz main beam brightness '
                                         'temperature: 1Hz'),
                           'units': 'K',
                           'valid_min': 135.15,
                           'valid_max': 265.15,
                           'options': {}}}

    idf_version = output_opts.get('idf_version', '1.0')
    granule = idf_converter.lib.create_granule(idf_version, DATA_MODEL)
    granule.vars = variables

    _input_path = input_opts.get('path', None)
    if _input_path is None:
        raise InputPathMissing()
    input_path = os.path.normpath(_input_path)

    freq_hz = int(input_opts.get('frequency', 1))

    if 1 == freq_hz:
        lon_name = 'lon_01'
        lat_name = 'lat_01'
        time_name = 'time_01'
        surface_name = 'surf_type_01'
        spatial_res_meters = 7000  # very approximative
        granule.vars['wind_speed_alt_01_ku'] = granule.vars.pop('wind')
        granule.vars['ssha_01_ku'] = granule.vars.pop('ssha')
        granule.vars['swh_ocean_01_ku'] = granule.vars.pop('swh')
        granule.vars['sig0_ocean_01_ku'] = granule.vars.pop('sig0')
        granule.vars['tb_238_01'] = granule.vars.pop('tb238')
        granule.vars['tb_365_01'] = granule.vars.pop('tb365')
        mask_methods = {'ssha_01_ku': mask_ssha,
                        'swh_ocean_01_ku': mask_swh,
                        'sig0_ocean_01_ku': mask_sig0,
                        'wind_speed_alt_01_ku': mask_wind}
    elif 20 == freq_hz:
        lon_name = 'lon_20_ku'
        lat_name = 'lat_20_ku'
        time_name = 'time_20_ku'
        surface_name = 'surf_type_20_ku'
        spatial_res_meters = 300  # very approximative
        del granule.vars['wind']  # no wind at 20Hz
        del granule.vars['tb238']  # no 23.8GHz brightness temperature at 20Hz
        del granule.vars['tb365']  # no 36.5GHz brightness temperature at 20Hz
        granule.vars['ssha_20_ku'] = granule.vars.pop('ssha')
        granule.vars['swh_ocean_20_ku'] = granule.vars.pop('swh')
        granule.vars['sig0_ocean_20_ku'] = granule.vars.pop('sig0')
        mask_methods = {'ssha_20_ku': mask_ssha,
                        'swh_ocean_20_ku': mask_swh,
                        'sig0_ocean_20_ku': mask_sig0}
    else:
        raise FrequencyNotSupported(freq_hz)

    channels = list(granule.vars.keys())

    f_handler = netCDF4.Dataset(input_path, 'r')
    _time = idf_converter.lib.extract_variable_values(f_handler, time_name)
    _lon = idf_converter.lib.extract_variable_values(f_handler, lon_name)
    _lat = idf_converter.lib.extract_variable_values(f_handler, lat_name)
    for var_id in granule.vars.keys():
        idf_converter.lib.extract_variable_attributes(f_handler, var_id,
                                                      granule)
        band = idf_converter.lib.extract_variable_values(f_handler, var_id)
        granule.vars[var_id]['array'] = band
        granule.vars[var_id]['datatype'] = band.dtype

    surface = idf_converter.lib.extract_variable_values(f_handler,
                                                        surface_name)
    tb238_quality = None
    if 'tb_238_01' in granule.vars.keys():
        flag_name = 'tb_238_quality_flag_01'
        tb238_quality = idf_converter.lib.extract_variable_values(f_handler,
                                                                  flag_name)
    tb365_quality = None
    if 'tb_365_01' in granule.vars.keys():
        flag_name = 'tb_365_quality_flag_01'
        tb365_quality = idf_converter.lib.extract_variable_values(f_handler,
                                                                  flag_name)

    idf_converter.lib.extract_global_attributes(f_handler, input_opts, granule)
    f_handler.close()

    granule.vars['lat'] = {'array': _lat,
                           'units': 'degrees north',
                           'datatype': _lat.dtype,
                           'options': {}}

    # Trick to deal with continuity in longitude
    lref = _lon[int(0.5 * numpy.shape(_lon)[0])]
    _lon = numpy.mod(_lon - (lref - 180.), 360.) + (lref - 180.)
    _lon = numpy.rad2deg(numpy.unwrap(numpy.deg2rad(_lon)))
    granule.vars['lon'] = {'array': _lon,
                           'units': 'degrees east',
                           'datatype': _lon.dtype,
                           'options': {}}

    # Apply time offset to transform time into an EPOCH timestamp
    ref_time = datetime.datetime(2000, 1, 1)
    epoch = datetime.datetime(1970, 1, 1)
    time_offset = (ref_time - epoch).total_seconds()
    shifted_time = _time + time_offset
    granule.vars['time'] = {'array': shifted_time.astype(numpy.double),
                            'units': 'seconds since 1970-01-01T00:00:00.000Z',
                            'datatype': numpy.double,
                            'options': {}}

    granule.dims['time'] = numpy.size(granule.vars['time']['array'])

    start_dt = datetime.datetime.utcfromtimestamp(shifted_time[0])
    stop_dt = datetime.datetime.utcfromtimestamp(shifted_time[-1])
    granule_name = os.path.basename(os.path.dirname(input_path))

    granule.meta['idf_granule_id'] = '{}_{:02d}Hz'.format(granule_name,
                                                          freq_hz)
    granule.meta['idf_subsampling_factor'] = 0
    granule.meta['idf_spatial_resolution'] = spatial_res_meters
    granule.meta['idf_spatial_resolution_units'] = 'm'
    granule.meta['time_coverage_start'] = start_dt
    granule.meta['time_coverage_end'] = stop_dt

    transforms = []

    transforms.append(('static_common_mask', {'targets': channels,
                                              'mask': (surface > 1)}))

    mwr_channels = [_ for _ in ('tb_238_01', 'tb_365_01',)
                    if _ in granule.vars.keys()]
    if 0 < len(mwr_channels):
        bands_mask = {'tb_238_01': (tb238_quality != 0),
                      'tb_365_01': (tb365_quality != 0)}
        transforms.append(('static_bands_mask', {'targets': mwr_channels,
                                                 'masks': bands_mask
                                                 }))
    else:
        logger.debug('No MWR variable')

    transforms.append(('remove_extra_lon_degrees', {'lon_name': 'lon'}))

    _channels = [_ for _ in channels if _ not in mwr_channels]
    transforms.append(('mask_methods', {'targets': _channels,
                                        'methods': mask_methods}))

    output_opts['__export'] = channels

    # data_model, dims, vars, attrs, formatter_jobs
    yield (input_opts, output_opts, granule, transforms)



# vim: ts=4:sts=4:sw=4
#
# @date 2019-09-25
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""This module defines constant values used in units conversion and data
reprojection.
"""

# Unit conversion
# 1deginmeter = 111.11 * 10**3
deginmeter = 111110
CelsiusinKelvin = 273.15

EPSG4326_WKT = 'GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563,AUTHORITY["EPSG","7030"]],AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0,AUTHORITY["EPSG","8901"]],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","4326"]]'  # noqa



# vim: ts=4:sts=4:sw=4
#
# @date 2019-09-25
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import numpy
import typing
import logging
import netCDF4
import datetime
from idf_converter.lib.types import InputOptions, OutputOptions, ProcessedLists
from idf_converter.lib.types import Granule, FormatterResult, GCPs

logger = logging.getLogger(__name__)


class MultipleDepthValues(Exception):
    """Error raised when the depth dimension is defined and its size is greater
    than 1: a single depth has to be selected in the reader!
    """
    pass


def help() -> typing.Tuple[str, str]:
    """"""
    inp = ('    geoloc_at_pixel_center\tBoolean set to true if the coordinates'
           ' provided in the input file correspond to the center of the pixels'
           ' and to false if they are referencing the top-left corner of the'
           ' pixels',
           )

    out = ('    gcp_spacing\tTODO',
           '    gcp_lat_spacing\tTODO',
           '    gcp_lon_spacing\tTODO',
           )
    return ('\n'.join(inp), '\n'.join(out))


def preformat(input_opts: InputOptions, output_opts: OutputOptions,
              granule: Granule) -> typing.Iterator[FormatterResult]:
    """"""
    if 'depth' in granule.dims.keys() and 'depth' in granule.vars.keys():
        if 1 < granule.dims['depth']:
            raise MultipleDepthValues()

        del granule.dims['depth']
        depth_var = granule.vars.pop('depth')
        depth = depth_var['array']
        granule_id = granule.meta['idf_granule_id']
        granule.meta['idf_granule_id'] = f'depth_{depth[0]}m_{granule_id}'

        granule.meta['geospatial_vertical_min'] = depth[0]
        granule.meta['geospatial_vertical_max'] = depth[0]
        if 'units' in depth_var.keys():
            depth_units = depth_var['units']
            granule.meta['geospatial_vertical_units'] = depth_units
        if 'positive' in depth_var.keys():
            depth_pos = depth_var['positive']
            granule.meta['geospatial_vertical_positive'] = depth_pos

    yield input_opts, output_opts, granule


def compute_gcps(input_opts: InputOptions,
                 output_opts: OutputOptions,
                 granule: Granule) -> GCPs:
    """"""
    # Number of digital points per GCP
    gcp_spacing = int(output_opts.get('gcp_spacing', 32))
    gcp_lat_spacing = int(output_opts.get('gcp_lat_spacing', gcp_spacing))
    gcp_lon_spacing = int(output_opts.get('gcp_lon_spacing', gcp_spacing))

    nlat = granule.dims['lat']
    nlon = granule.dims['lon']

    # Compute GCPs
    lon = granule.vars['lon']['array']
    lat = granule.vars['lat']['array']

    # Handle longitude continuity
    dlon = lon[1:] - lon[:-1]
    if 180.0 <= numpy.max(numpy.abs(dlon)):
        lon_neg_ind = numpy.where(lon < 0.0)
        lon[lon_neg_ind] = lon[lon_neg_ind] + 360.0

    # Compute geotransform parameters
    # A workaround is used here to avoid numerical precision issues in
    # numpy.mean: if values are too close to 0, underflow errors may arise so
    # we multiply values by a large factor before passing them to numpy.mean,
    # then we divide the result by the same factor
    precision_factor = 10000
    lon0 = lon[0]
    _dlon = lon[1:] - lon[:-1]
    try:
        dlon = numpy.mean(_dlon)
    except FloatingPointError:
        dlon = numpy.mean(precision_factor * _dlon) / precision_factor
    lat0 = lat[0]
    _dlat = lat[1:] - lat[:-1]
    try:
        dlat = numpy.mean(_dlat)
    except FloatingPointError:
        dlat = numpy.mean(precision_factor * _dlat) / precision_factor
    x0, dxx, dxy, y0, dyx, dyy = [lon0, dlon, 0, lat0, 0, dlat]

    logger.debug(f'Geotransform: {x0} {dxx} {dxy} {y0} {dyx} {dyy}')

    # Compute number of GCPs (except the ones for bottom and right edge)
    # according to the requested resolution, i.e. the number of digital points
    # between two GCPs
    gcp_nlat = numpy.ceil(nlat / gcp_lat_spacing).astype('int')
    gcp_nlon = numpy.ceil(nlon / gcp_lon_spacing).astype('int')

    logger.debug(f'{nlat}, {nlon}, {gcp_lat_spacing}, {gcp_lon_spacing},'
                 f'{gcp_nlat}, {gcp_nlon}')

    # Compute matrix indices for the GCPs
    gcp_lin = numpy.arange(gcp_nlat) * gcp_lat_spacing
    gcp_pix = numpy.arange(gcp_nlon) * gcp_lon_spacing

    # Add an extra line and column to geolocate the bottom and right edges of
    # the data matrix
    gcp_lin = numpy.concatenate((gcp_lin, [nlat]))
    gcp_pix = numpy.concatenate((gcp_pix, [nlon]))

    # GCP pixels are located at the edge of data pixels, meaning that the
    # center of the first data pixel is located at (0.5, 0.5) in the GCPs
    # matrix.
    # If lon/lat correspond to the center of data pixels, then the indices
    # stored in the GCPs must be shifted by one half-pixel along each axis
    _geoloc_at_pixel_center = input_opts.get('geoloc_at_pixel_center', 'true')
    geoloc_at_pixel_center = _geoloc_at_pixel_center.lower() in ('true', 'yes')
    if geoloc_at_pixel_center:
        logger.debug('Shift GCPs by ½ pixel')
        x0 = x0 - 0.5 * dxx
        y0 = y0 - 0.5 * dyy

    # Compute GCP geographical coordinates expressed in lat/lon
    _gcp_lin = gcp_lin[:, numpy.newaxis]
    _gcp_pix = gcp_pix[numpy.newaxis, :]
    gcp_lat = y0 + dyx * _gcp_pix + dyy * _gcp_lin
    gcp_lon = x0 + dxx * _gcp_pix + dxy * _gcp_lin

    return gcp_lon, gcp_lat, gcp_lin, gcp_pix


def format_granule(input_opts: InputOptions,
                   output_opts: OutputOptions,
                   granule: Granule,
                   idf_file: netCDF4.Dataset) -> ProcessedLists:
    """Create dimensions, variables and attributes required for swath granules.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format
    granule: idf_converter.lib.Granule
        Object representing the granule data and metadata
    idf_file: netCDF4.Dataset
        Handle of the IDF file to edit

    Returns
    -------
    tuple
        Information about granule information that the method has processed.
        The tuple contains three lists::
        - a list of names for the dimensions created by the method
        - a list of identifiers for the variables created by the method
        - a list of names for the attributes created by the method
    """

    gcp_lon, gcp_lat, gcp_lin, gcp_pix = compute_gcps(input_opts, output_opts,
                                                      granule)
    gcp_nlat, gcp_nlon = gcp_lin.size, gcp_pix.size

    nlat = granule.dims['lat']
    nlon = granule.dims['lon']

    # Central datetime
    _start_dt = granule.meta['time_coverage_start']
    _stop_dt = granule.meta['time_coverage_end']
    mid_time = _start_dt + (_stop_dt - _start_dt) / 2

    # Precision set to seconds
    microseconds = mid_time.microsecond
    if microseconds > 500000:
        mid_time = mid_time + datetime.timedelta(seconds=1)
    mid_time = mid_time.replace(microsecond=0)

    # Dimensions
    idf_file.createDimension('time', None)  # unlimited
    idf_file.createDimension('lat', nlat)
    idf_file.createDimension('lon', nlon)
    idf_file.createDimension('lat_gcp', gcp_nlat)
    idf_file.createDimension('lon_gcp', gcp_nlon)
    idf_dims = ('time', 'lat', 'lon', 'lat_gcp', 'lon_gcp',)
    extra_dims = [x for x in granule.dims.keys() if x not in idf_dims]
    for extra_dim in extra_dims:
        idf_file.createDimension(extra_dim, granule.dims[extra_dim])

    # Variables
    _time = idf_file.createVariable('time', 'f8', ('time', ))
    _time.long_name = 'time'
    _time.standard_name = 'time'
    _time.units = 'seconds since 1970-01-01T00:00:00.000000Z'
    _time.calendar = 'standard'
    _time[:] = (mid_time - datetime.datetime(1970, 1, 1)).total_seconds()

    _latgcp = idf_file.createVariable('lat_gcp', 'f4', ('lat_gcp',))
    _latgcp.long_name = 'ground control points latitude'
    _latgcp.standard_name = 'latitude'
    _latgcp.units = 'degrees_north'
    _latgcp[:] = gcp_lat[:, 0].astype('float32')

    _longcp = idf_file.createVariable('lon_gcp', 'f4', ('lon_gcp',))
    _longcp.long_name = 'ground control points longitude'
    _longcp.standard_name = 'longitude'
    _longcp.units = 'degrees_east'
    _longcp[:] = gcp_lon[0, :].astype('float32')

    _indexlatgcp = idf_file.createVariable('index_lat_gcp', 'i4', ('lat_gcp',))
    _indexlatgcp.long_name = ('index of ground control points in lat '
                              'dimension')
    _indexlatgcp.comment = ('index goes from 0 (start of first pixel) to '
                            'dimension value (end of last pixel)')
    _indexlatgcp[:] = gcp_lin.astype('int32')

    _indexlongcp = idf_file.createVariable('index_lon_gcp', 'i4', ('lon_gcp',))
    _indexlongcp.long_name = 'index of ground control points in lon dimension'
    _indexlongcp.comment = ('index goes from 0 (start of first pixel) to '
                            'dimension value (end of last pixel)')
    _indexlongcp[:] = gcp_pix.astype('int32')

    idf_file.cdm_data_type = b'Grid'

    processed_dims = ['time', 'lat', 'lon', 'lat_gcp', 'lon_gcp']
    processed_vars = ['time', 'lat_gcp', 'lon_gcp', 'index_lat_gcp',
                      'index_lon_gcp', 'lat', 'lon']
    processed_attrs = ['cdm_data_type']

    return (processed_dims, processed_vars, processed_attrs)



# vim: ts=4:sts=4:sw=4
#
# @date 2019-09-27
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import os
import numpy
import typing
import netCDF4
import logging
import datetime
import idf_converter.lib
from idf_converter.lib.types import InputOptions, OutputOptions, ReaderResult
from idf_converter.lib.types import TransformsList

logger = logging.getLogger(__name__)

DATA_MODEL = 'SWATH'


class InputPathMissing(IOError):
    """"""
    pass


class MissingVariablesList(Exception):
    """Error raised when the "variables" input option has not been specified
    by the user."""
    pass


class MaskedCoordinatesError(Exception):
    """Error raised when masked values are found in longitudes or latitudes
    even after removing the leading and trailing rows where coordinates are
    completely masked."""
    pass


def help() -> typing.Tuple[str, str]:
    """Describe options supported by this reader.

    Returns
    -------
    str
        Description of the supported options
    """
    inp = ('    path\tPath of the input file',)
    out = ('',)
    return ('\n'.join(inp), '\n'.join(out))


def read_data(input_opts: InputOptions,
              output_opts: OutputOptions
              ) -> typing.Iterator[ReaderResult]:
    """Read input file, extract data and metadata, store them in a Granule
    object then prepare formatting instructions to finalize the conversion to
    IDF format.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule, list)
        A tuple which contains four elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the extracted information
          has been stored

        - a :obj:list of :obj:dict describing the formatting operations that
          the converter must perform before serializing the result in IDF
          format
    """

    idf_version = output_opts.get('idf_version', '1.0')
    granule = idf_converter.lib.create_granule(idf_version, DATA_MODEL)
    idf_converter.lib.apply_global_overrides(input_opts, granule)
    idf_converter.lib.apply_var_overrides(input_opts, granule)

    _input_path = input_opts.get('path', None)
    if _input_path is None:
        raise InputPathMissing()
    _var_ids = input_opts.get('variables', None)
    if _var_ids is None:
        raise MissingVariablesList()
    var_ids = [x.strip() for x in _var_ids.split(',')]

    extract_variable_values = idf_converter.lib.extract_variable_values
    input_path = os.path.normpath(_input_path)
    f_handler = netCDF4.Dataset(input_path, 'r')
    lon = extract_variable_values(f_handler, 'lon')
    lat = extract_variable_values(f_handler, 'lat')
    idf_converter.lib.extract_global_attributes(f_handler, input_opts, granule)

    final_shape = numpy.shape(lon)
    track_slice = slice(None, None, None)
    scan_slice = slice(None, None, None)
    time_slice = slice(0, 1, None)

    lon_mask = numpy.ma.getmaskarray(lon)
    lat_mask = numpy.ma.getmaskarray(lat)
    coords_masked = (lon_mask | lat_mask)
    if coords_masked.all():
        logger.warning('All coordinates are masked. Skipping this file.')
        raise idf_converter.lib.EarlyExit()

    # Compute alongtrack slice after removing leading and trailing rows that
    # contain only masked coordinates
    rows_mask = numpy.logical_and.reduce(coords_masked, axis=1)
    valid_rows_ind = numpy.where(~rows_mask)
    track_slice = slice(valid_rows_ind[0][0], 1 + valid_rows_ind[0][-1])
    data_slices = [time_slice, track_slice, scan_slice]
    if 0 < track_slice.start:
        msg = 'Coordinates were masked in the {} first rows'
        logger.info(msg.format(track_slice.start))
    if track_slice.stop < final_shape[0]:
        msg = 'Coordinates were masked in the {} last rows'
        logger.info(msg.format(final_shape[0] - track_slice.stop))

    # Remove rows without coordinates from lon, lat and quality mask
    lon = lon[track_slice, :]
    lat = lat[track_slice, :]
    final_shape = numpy.shape(lon)

    # Check that there are no masked coordinates anymore
    lon_mask = numpy.ma.getmaskarray(lon)
    lat_mask = numpy.ma.getmaskarray(lat)
    coords_masked = (lon_mask | lat_mask)
    if coords_masked.any():
        raise MaskedCoordinatesError()

    masked_var_count = 0
    for var_id in var_ids:
        idf_converter.lib.extract_variable_attributes(f_handler, var_id,
                                                      granule)
        if var_id in f_handler.variables.keys():
            band = extract_variable_values(f_handler, var_id, idx=data_slices)
        else:
            # MODIS A files are not homogenous: some files contain
            # chlorophyll_a and K_490 whereas other files don't. IDF granule
            # *must* be homogenous so we create these variables and assign
            # them a fully masked array.
            msg = 'Variable {} is missing, a placeholder is used instead'
            logger.warning(msg.format(var_id))
            band = numpy.ma.masked_all(final_shape, float)

        # Apply flag mask
        flag_var_id = granule.vars[var_id].get('flag_variable', None)
        if flag_var_id is not None:
            flag_band = f_handler.variables[flag_var_id][data_slices]
            mask = idf_converter.lib.build_flag_mask(var_id, band, flag_band,
                                                     granule)
            # Transform mask so that it has the same number of dimensions as
            # the data band, assuming data band dimensions come last in the
            # dimensions list of the mask
            while len(mask.shape) > len(band.shape):
                mask = mask[0]
            band = numpy.ma.masked_where(mask, band)

        band_masked = False
        if numpy.ma.getmaskarray(band).all():
            band_masked = True
            masked_var_count = masked_var_count + 1

        granule.vars[var_id]['array'] = band
        granule.vars[var_id]['datatype'] = band.dtype
        if 'name' not in granule.vars[var_id].keys():
            granule.vars[var_id]['name'] = var_id

        finite_ind = numpy.where(numpy.isfinite(band))
        if granule.vars[var_id].get('valid_min', None) in (None, 'auto'):
            vmin = 0.0
            if band_masked:
                logger.warning('Impossible to compute a valid_min '
                               f'automatically for {var_id} because all the '
                               f'values are masked. Using {vmin} by default.')
            else:
                vmin = numpy.nanmin(band[finite_ind])
            granule.vars[var_id]['valid_min'] = vmin

        if granule.vars[var_id].get('valid_max', None) in (None, 'auto'):
            vmax = 1.0
            if band_masked:
                logger.warning('Impossible to compute a valid_max '
                               f'automatically for {var_id} because all the '
                               f'values are masked. Using {vmax} by default.')
            else:
                vmax = numpy.nanmax(band[finite_ind])
            granule.vars[var_id]['valid_max'] = vmax

    if masked_var_count >= len(var_ids):
        # All variables are fully masked, exit early to avoid the creation of
        # a useless IDF file
        raise idf_converter.lib.EarlyExit()

    nrow, ncell = numpy.shape(lon)
    start_dt = datetime.datetime.strptime(f_handler.start_time,
                                          '%Y%m%dT%H%M%SZ')
    stop_dt = datetime.datetime.strptime(f_handler.stop_time, '%Y%m%dT%H%M%SZ')
    f_handler.close()

    granule_file = os.path.basename(input_path)
    granule_name, _ = os.path.splitext(granule_file)

    granule.meta['idf_subsampling_factor'] = 0
    granule.meta['idf_spatial_resolution'] = 1000
    granule.meta['idf_spatial_resolution_units'] = 'm'
    granule.meta['idf_granule_id'] = f'{granule_name}'
    granule.dims['row'] = int(nrow)
    granule.dims['cell'] = int(ncell)

    # Format spatial information
    granule.vars['lat'] = {'array': lat,
                           'units': 'degrees_north',
                           'datatype': lat.dtype,
                           'options': {}}
    granule.vars['lon'] = {'array': lon,
                           'units': 'degrees_east',
                           'datatype': lon.dtype,
                           'options': {}}

    # Format time information
    granule.meta['time_coverage_start'] = start_dt
    granule.meta['time_coverage_end'] = stop_dt

    # Add transforms
    transforms: TransformsList = []

    output_opts['__export'] = var_ids

    yield (input_opts, output_opts, granule, transforms)
